#include "Action.h"
#include "Configuration.h"

#include "../worldmodel/WorldModelInterface.h"

/**
 * Aktueller Stand
 *
 *   $RCSfile: Action.cc,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 2001/07/31 22:23:08 $
 *   $Author: rami $
 *
 */

Action::Action()
  : theActionId( UNDEF_ACTION ),
    theSimStep( UNDEF_INT ),
    theActionCount( UNDEF_INT )
{
}

Action::Action( int actionId, int simStep, double firstParameter, double secondParameter )
  : theActionId( actionId ),
    theSimStep( simStep ),
    theActionCount( UNDEF_INT ),
    theFirstParameter( firstParameter ),
    theSecondParameter( secondParameter )
{
  if( actionId < MIN_DEFAULT_ACTION || actionId > MAX_DEFAULT_ACTION ) {
    throwException( ActionException( "Wrong actionId: " + actionId ) );
  }
}

Action::Action( int actionId, int simStep, int viewWidth, int viewQuality )
  : theActionId( actionId ),
    theSimStep( simStep ),
    theActionCount( UNDEF_INT ),
    theViewWidth( viewWidth ),
    theViewQuality( viewQuality )
{
  if( actionId != CHANGE_VIEW_ACTION ) {
    throwException( ActionException( "Wrong actionId: " + actionId ) );
  }
}

Action::Action( int actionId, int simStep, const Ath98String& message )
  : theActionId( actionId ),
    theSimStep( simStep ),
    theActionCount( UNDEF_INT ),
    theMessage( message )
{
  if( actionId != SAY_ACTION ) {
    throwException( ActionException( "Wrong actionId: " + actionId ) );
  }
}

Action::~Action()
{
}

bool Action::isValid() const {
  switch( theActionId ) {
  case MOVE_ACTION:
    if ( theFirstParameter + DEFAULT_COMPARISON_ACCURACY < -60.0 ||
        theFirstParameter - DEFAULT_COMPARISON_ACCURACY > 60.0 ||
        theSecondParameter + DEFAULT_COMPARISON_ACCURACY < -37.0 ||
        theSecondParameter - DEFAULT_COMPARISON_ACCURACY > 37.0 ) {
      return NO;
    }
    break;
  case KICK_ACTION:
    if( theFirstParameter + DEFAULT_COMPARISON_ACCURACY < -30.0 ||
	theFirstParameter - DEFAULT_COMPARISON_ACCURACY > 100.0 ) {
      return NO;
    }
    break;
  case DASH_ACTION:
    if( theFirstParameter + DEFAULT_COMPARISON_ACCURACY < -100.0 ||
	theFirstParameter - DEFAULT_COMPARISON_ACCURACY > 100.0 ) {
      return NO;
    }
    break;
  case TURN_ACTION:
  case TURN_NECK_ACTION:
  case CATCH_ACTION:
  case SENSE_BODY_ACTION:
  case EMPTY_ACTION:
    break;
  case SAY_ACTION:
    if( theMessage.length() + 1 >= globalConf->sayMessageSize ) {
      return NO;
    }
    break;
  case CHANGE_VIEW_ACTION:
    switch( theViewWidth) {
    case VIEW_WIDTH_NARROW:
    case VIEW_WIDTH_NORMAL:
    case VIEW_WIDTH_WIDE:
      break;
    default:
      return NO;
    }
    switch( theViewQuality) {
    case VIEW_QUALITY_LOW: 
    case VIEW_QUALITY_HIGH: 
      break;  
    default: 
      return NO;
    }
    break;
  default:
    return NO;
  }
  return YES;
}

bool Action::isMainAction() const {
  switch( theActionId ) {
    case SENSE_BODY_ACTION:
    case SAY_ACTION:
    case TURN_NECK_ACTION:
    case CHANGE_VIEW_ACTION:
      return NO;
      break;
    default:
      return YES;
  }
}

double Action::correctAngleToServerSystem(double anAngle){
  double angle = -1.0 * normalizeAngle(anAngle);
  if( fabs(angle) > Ath98_PI )
    angle += 2.0 * Ath98_PI;
  return angle;
}

Ath98String Action::asCommandAth98String() const {
  Ath98String viewWidth;
  Ath98String viewQuality;

  switch( theActionId ) {
  case MOVE_ACTION:
    return( "(move " + Ath98String(theFirstParameter) + " " +
	    Ath98String(-theSecondParameter) + ")" );
    break;
  case KICK_ACTION:
    return( "(kick " + Ath98String(theFirstParameter) + " " +
	    Ath98String(degFromRadiant(correctAngleToServerSystem(theSecondParameter))) + ")" );
    break;
  case DASH_ACTION:
    return( "(dash " + Ath98String(theFirstParameter) + ")" );
    break;
  case TURN_ACTION:
    return( "(turn " + Ath98String(degFromRadiant(correctAngleToServerSystem(theFirstParameter))) + ")" );
    break;
  case TURN_NECK_ACTION:
    return( "(turn_neck " + Ath98String(degFromRadiant(correctAngleToServerSystem(theFirstParameter))) + ")" );
    break;
  case CATCH_ACTION:
    return( "(catch " + Ath98String(degFromRadiant(correctAngleToServerSystem(theFirstParameter))) + ")" );
    break;
  case SENSE_BODY_ACTION:
    return( "(sense_body)" );
    break;
  case EMPTY_ACTION:
    return Ath98String();
    break;
  case SAY_ACTION:
    return( "(say " + theMessage + ")" );
    break;
  case CHANGE_VIEW_ACTION:
    switch( theViewWidth ) {
    case VIEW_WIDTH_NARROW:
      viewWidth = "narrow";
      break;
    case VIEW_WIDTH_NORMAL:
      viewWidth = "normal";
      break;
    case VIEW_WIDTH_WIDE:
      viewWidth = "wide";
      break;
    }
    switch( theViewQuality ) {
    case VIEW_QUALITY_LOW:
      viewQuality = "low";
      break;
    case VIEW_QUALITY_HIGH:
      viewQuality = "high";
      break;
    }
    return( "(change_view " + viewWidth + " " + viewQuality + ")" );
    break;
  default:
    throwException(
        ActionException("Invalid Action with id " + Ath98String( theActionId ) ) );
    return Ath98String();
  }
}

string Action::asCommandString() const{
  return string( asCommandAth98String().cString() );
}

Action* Action::copy() const {
  return new Action( *this );
}

Ath98String Action::toString() const
{
  Ath98String output = Ath98String( "( Action for simStep " + Ath98String( simStep() ) );
  if ( id() == EMPTY_ACTION ) {
    output = output + " none )";
  }
  else {
    output = output + Ath98String( " message: " ) + asCommandAth98String() + " )";
  }
  return output;
}

string Action::convertMessage( const Ath98String& aMessage ) const{
  char message[MAX_MESSAGE];
  char encodedSenderinfo[MAX_MESSAGE];
  long senderinfo;
  char simstep[3];

  sscanf( aMessage.cString(), "%ld %s %s %[^\n]", &senderinfo,
	  encodedSenderinfo, simstep, message );
  string result( message );
  return result;
}
  
