#ifndef _ACTION_H_
#define _ACTION_H_

#include "Common.h"
#include "Ath98String.h"
#include "Exception.h"
#include <stream.h>
#include <string>
/**
  Elementare Handlung.
  Entspricht einem Kommando, das an den SoccerServer geschickt werden kann.
  Zur selben Zeit koennen mehrere Aktionen ausgefuehrt werden.
  Jede Aktion hat auch einen simStep, f\"ur den sie berechnet wurde. Stimmt der
  aktuelle simStep beim Abschicken mit dem SimStep der Aktion nicht \"uberein,
  so liegt ein Synchronisationsfehler vor.

  Aktionen beliebigen Typs von dieser Klasse gehandhabt werden. Dadurch gibt es
  mehrere Konstruktoren. Einige Instanzvariablen sind immer undefiniert (die
  fuer die Aktion, die gerade gespeichert ist, nicht ben\"otigt werden).

  {\small {\bf Quelltext}

  [\URL[Header]{src/Action.h} | [\URL[Implementierung]{src/Action.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Action.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:08 $
  *   $Author: rami $
  *
  **************************************************
  \end{verbatim} }*/
class Action
{
public:

  /** @name Fehlerklassen */
  //@{
    /** Allgemeine Actionfehler-Klasse */
    class ActionException: public Exception {
    public:
      ActionException(const Ath98String& aDescription )
        : Exception( aDescription ) {}
    };
  //@}


  /** @name Objekt erzeugen und zerst\"oren */
  //@{
    /** Default-Konstruktor.
      Erzeugt eine UNDEF_ACTION. */
    Action();

    /** Konstruktor fuer die Aktionen dash, turn, kick
      move, catch, senseBody und emptyAction. Die Parameter haben folgende
      Bedeutung:
      \begin{description}
        \item[dash]firstParameter: Kraft, secondParameter: -
        \item[kick]firstParameter: Kraft, secondParameter: Kickrichtung
        \item[turn]firstParameter: Drehwinkel, secondParameter: -
        \item[catch]firstParameter: Hechtrichtung, secondParameter: -
        \item[move]firstParameter: x-Koordinate, secondParameter:
          y-Koordinate
        \item[senseBody]firstParameter: -, secondParameter: -
        \item[emptyAction]firstParameter: -, secondParameter: -
      \end{description}

      @param actionId Die Id der Aktion.
      @param simStep der simStep, fuer den die Aktion berechnet ist.
      @param firstParameter  
        Der erste Parameter der Aktion. Falls die Aktion keinen ersten
        Parameter dieses Typs benoetigt, ist der Wert undefiniert. Fuer die
        Aktionen dash, turn, kick, move und catch ist dieser Parameter noetig.
      @param secondParameter
        Der zweite Parameter der Aktion. Falls diese Aktion keinen zweiten
        Parameter dieses Typs benoetigt, ist der Wert undefiniert. Fuer die
        Aktionen kick, und move ist dieser Parameter noetig.
      @see Aktionen */
    Action(
      int actionId,
      int simStep,
      double firstParameter = UNDEF_DOUBLE,
      double secondParameter = UNDEF_DOUBLE );

    /** Konstruktor fuer die Aktion changeView.
      @param actionId Die Id der Aktion. Der Wert {\em mu\3} CHANGE_VIEW_ACTION
        sein.
      @param simStep der simStep, fuer den die Aktion berechnet ist.
      @param viewWidth Der gewuenschte Sichtwinkel der Aktion. Dieser kann
        folgende Werte annehmen: VIEW_WIDTH_NARROW, VIEW_WIDTH_NORMAL,
        VIEW_WIDTH_WIDE.
      @param viewQuality Die gewuenschte Sichtqualitaet der Aktion. Diese kann
        folgende Werte annehmen: VIEW_QUALITY_LOW, VIEW_QUALITY_HIGH.
      @see Aktionen */
    Action( int actionId, int simStep, int viewWidth, int viewQuality );

    /** Konstrukter fuer die Aktion say.
      @param actionId Die Id der Aktion. Diese {\em mu\3} SAY_ACTION sein.
      @param simStep der simStep, fuer den die Aktion berechnet ist.
      @param message Die Nachricht der say Aktion. Diese kann ein
      beliebiger Ath98String nach den SoccerServer Konventionen sein. */
    Action( int actionId, int simStep, const Ath98String& message );

    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
    Action* copy() const;

    /** Der Destruktor */
    ~Action();

  //@}


  /** @name Abfrage und Setzen von Daten */
  //@{

    /** Liefert die Id der Aktion.
      @see Aktionen */
    int id() const;

    /** Liefert den simStep der Aktion. */
    int simStep() const;

    /** Liefert die Anzahl der bereits vorher abgesendeten Aktionen
      dieses Typs.
      @return UNDEF_INT wenn die Anzahl nicht bekannt ist sonst die
      Anzahl */
    int actionCount() const;

    /** Liefert den ersten Parameter.
      Hat die Aktion keinen Parameter, so wird UNDEF_DOUBLE geliefert. */
    double firstParameter() const;

    /** Liefert den zweiten Parameter.
      Hat die Aktion keine zwei Parameter, so wird UNDEF_DOUBLE geliefert. */
    double secondParameter() const;

    /** Liefert die Sichtbreite.
      Wenn es sich nicht um eine CHANGE_VIEW_ACTION handelt, wird
      UNDEF_INT geliefert. */
    int viewWidth() const;

    /** Ueberprueft, ob alle f\"ur eine Aktion des Typs \Ref{id}() notwendigen
      Parameter vorhanden und in einem gueltigen Bereich sind. */
    bool isValid() const;

    /** Liefert zurueck, ob die Aktion eine Hauptaktion ist.
      Pro Takt wird vom Server nur eine Hauptaktion ausgefuehrt. Hauptaktionen
      sind da */
    bool isMainAction() const;

    /** Setzt den Simulationsschritt, zu dem die Aktion ausgef"uhrt wird.
      Das nachtr\"agliche Umsetzen ist notwendig, damit bei der Registrierung
      im Weltmodell bekannt ist, zu welchem Zeitpunkt die Aktion wirklich
      gesendet wurde. Das ist nicht notwendigerweise der Zeitpunkt, f\"ur den
      die Aktion gedacht war.
      @param newSimStep der neue Simulationsschritt der Aktion */
    void setSimStep( const int newSimStep );

    /** Setzt die Anzahl, der bereits vor dieser Aktion abgesendeten
      Aktionen.
      @param vorher abgesendete Aktionen dieses Typs */
    void setActionCount( const int aActionCount );
  //@}


  /** @name Umwandlungen in Zeichenketten */
  //@{
    /** Liefert einen Ath98String, der, an den SoccerServer gesandt, die Aktion
      ausloest.
      Dabei werden vorher evtl. notwendige Koordinatentransformationen
      durchgefuehrt. */
    Ath98String asCommandAth98String() const;

    /** Liefert einen STL-String, der, an den SoccerServer gesandt,
	die Aktion ausl\"ost.
	Dabei werden vorher evtl. notwendige Koordinatentransformationen
	durchgefuehrt. */
    string asCommandString() const;

    /** berfhrt unser Winkelsystem in das vom SoccerServer. */
    double correctAngleToServerSystem(double anAngle);


    /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal. */
    Ath98String toString() const;
  //@}

  /** Ausgabe auf StampStream */
  void toStampStream( ostream& ) const;

  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const Action& );

private:
  /** Die Id der Aktion. */
  int theActionId;

  /** Der simStep, fuer den die Aktion gemeint ist */
  int theSimStep;

  /** Anzahl der bereits vorher abgesendeten Aktionen dieses Typs. */
  int theActionCount;

  /** Der erste Parameter der Aktion. Falls diese Aktion keinen ersten
    Parameter dieses Typs benoetigt, ist der Wert undefiniert. Fuer die
    Aktionen dash, turn, kick, move und catch ist dieser Parameter noetig. */
  double theFirstParameter;

  /** Der zweite Parameter der Aktion. Falls diese Aktion keinen zweiten
    Parameter dieses Typs benoetigt, ist der Wert undefiniert. Fuer die
    Aktionen kick, und move ist dieser Parameter noetig. */
  double theSecondParameter;

  /** Der gewuenschte Sichtwinkel fuer die change_view Aktion. Dieser kann
    folgende Werte annehmen: VIEW_WIDTH_NARROW, VIEW_WIDTH_NORMAL,
    VIEW_WIDTH_WIDE. */
  int theViewWidth;

  /** Die gewuenschte Sichtqualitaet fuer die change_view Aktion. Diese kann
    folgende Werte annehmen: VIEW_QUALITY_LOW, VIEW_QUALITY_HIGH. */
  int theViewQuality;

  /** Die Nachricht bei einer say Aktion. Diese kann einen beliebigen Ath98String
    nach den SoccerServer Konventionen enthalten. */
  Ath98String theMessage;

  /** Wandelt die Say-Message in einen Lesbaren String um, damit dieser
      auf den Stamp-Stream ausgegeben werden kann. */
  string convertMessage( const Ath98String& aMessage ) const;

};

inline int Action::id() const
{
  return theActionId;
}

inline double Action::firstParameter() const
{
  return theFirstParameter;
}

inline double Action::secondParameter() const
{
  return theSecondParameter;
}

inline int Action::viewWidth() const
{
  if( theActionId == CHANGE_VIEW_ACTION ) {
    return theViewWidth;
  }
  else {
    return UNDEF_INT;
  }
}

inline int Action::simStep() const
{
  return theSimStep;
}

inline void Action::setSimStep( const int newSimStep )
{
  theSimStep = newSimStep;
}

inline int Action::actionCount() const
{
  return theActionCount;
}

inline void Action::setActionCount( const int aActionCount ) {
  theActionCount = aActionCount;
}


inline ostream& operator<< (ostream& os, const Action& a )
{
  os << "( Action for simStep " << a.simStep();

  if ( a.id() == EMPTY_ACTION ) {
    os  << " none )";
  }
  else {
    os << " message: "  <<  a.asCommandAth98String() + " )";
  }
  return os;

}
inline void Action::toStampStream( ostream& os) const {
  os <<
    " " << simStep() <<
    " " << id();

  switch( theActionId ) {
  case MOVE_ACTION:
    os << " " << theFirstParameter << " " << -theSecondParameter;
    break;
  case KICK_ACTION:
    os << " " << theFirstParameter << " " << -degFromRadiant(theSecondParameter);
    break;
  case DASH_ACTION:
    os << " " << theFirstParameter;
    break;
  case TURN_ACTION:
    os << " " << -degFromRadiant(theFirstParameter);
    break;
  case TURN_NECK_ACTION:
    os << " " << -degFromRadiant(theFirstParameter);
    break;
  case CATCH_ACTION:
    os << " " << -degFromRadiant(theFirstParameter);
    break;
  case SENSE_BODY_ACTION:
    break;
  case EMPTY_ACTION:
    break;
  case SAY_ACTION:
    os << " " << theMessage.length() << " " << convertMessage( theMessage )
       << SAY_MESSAGE_END_MARK;
    break;
  case CHANGE_VIEW_ACTION:
    os << " " << theViewWidth << " " << theViewQuality;
    break;
  default:
    throwException(
        ActionException("Invalid Action with id " + Ath98String( theActionId ) ) );
  }
}

#endif
