#ifndef _TIMER_H_
#define _TIMER_H_

#include "Deliberator.h"
#include "Exception.h"
#include "Say.h"
#include "../serverInterface/BodyInfos.h"
#include "../serverInterface/Communicator.h"
#include "../serverInterface/Effectors.h"
#include "../serverInterface/Sensors.h"
#include "../worldmodel/Neck.h"

/**
   Diese Klasse regelt das Echtzeitverhalten des Agenten.
   Hierzu ruft der Deliberator sehr haeufig die \ref{controller}-Funktion auf,
   die kurz prueft, ob sie eine Aufgabe hat und den Deliberator
   weiterarbeiten laesst, wenn dies nicht der Fall ist. Ihre beiden
   Aufgaben sind:
   1) Einlesen der Daten vom SoccerServer durch Aufruf entsprechender
   Klassen (Sensoren, Weltmodell)
   2) zeitlich optimales Abschicken der Aktionen an den SoccerServer
   durch Aufruf entspechender Klassen (Effektoren, PSI, Neck, Say-Skill)
 
   Verbesserungsmoeglichkeit:
   Optimierung der Zeiten durch Pruefung, ob abgeschickte Kommandos
   rechtzeitig angekommen sind.
 
   (diese Verbesserung sollte erst erfolgen, wenn die Klasse
   funktionsfaehig implementiert ist)
 
  {\small {\bf Quelltext}

  [\URL[Header]{src/Timer.h} | [\URL[Implementierung]{src/Timer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Timer.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:08 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/


class Timer
{
 public:

  /** @name Fehlerklassen */
  //@{
    /** Generelle Timerfehler-Klasse. */
/*
    class TimerException: public Exception {
    public:
      TimerException(const Ath98String& aDescription )
        : Exception( aDescription ) {}
    };
*/
  //@}
  
  /** @name Timer erzeugen und entfernen */
  //@{

    /** Erzeugt einen neuen Timer. 
       Dieser ist noch nicht aktiv.Er wird erst aktiv, wenn er mit run() 
       gestartet wird.
       @param sensors Zeiger auf das Sensorik-Objekt.
       @param effectors Zeiger auf das Effector-Objekt.
       @param deliberator Zeiger auf das Deliberator-Objekt. */
   Timer( Sensors *sensors, Effectors *effectors, Deliberator *deliberator, 
          Neck *neck, Say *say );


    
    ~Timer();
  
  //@}
    
  /** Testet auf Zeit/Sensorinformation/usw. und generiert gegebenenfalls
    Aktion und schickt sie an den Server oder integriert Sensorinfos
    ins Weltmodell
	  
    @return YES oder NO
    YES : falls neue Informationen da sind, oder eine Action gesendet wurde
    NO : sonst. */
  bool controller();

  /** Funktion, die im playSoccerIntention->execute() aufgerufen wird 
      falls das execute moeglicherweise zu lange dauert
      Es ist dazu da dem Timer die Gelegenheit zu geben doch noch irgendeine
      Aktion zu senden.
      Falls YES zurueckgegeben wird soll die Aktion zurueck geben werden,
      die das execute() der EmergencyIntention liefert; anderenfalls ist
      noch Zeit um execute/repair weiterarbeiten zu lassen


      @return YES oder NO
      YES: wenn das Abschicken einer billigen Aktion verursacht, durch die
           EmergencyIntention ausgeloest werden soll, weil es schon spaet ist.
      NO: wenn das execute() weitermachen soll */
  bool isEmergency() const;

  /** Folgende Funktionen liefern dem Deliberator Informationen, was
      der controller gemacht hat, wenn dessen Ausgabe YES war*/

  /** @return YES oder NO
    YES : wenn MainAction gesendet wurde
    NO : sonst */
  bool isMainActionSent();

 /** @return YES oder NO
    YES : wenn ViewAction gesendet wurde
    NO : sonst */
  bool isViewActionSent();

 /** @return YES oder NO
    YES : wenn NeckAction gesendet wurde
    NO : sonst */
  bool isNeckActionSent(); 

 /** @return YES oder NO
    YES : wenn SayAction gesendet wurde
    NO : sonst */
  bool isSayActionSent(); 
  
  /** @return YES oder NO
    YES : wenn irgendeine Aktion gesendet wurde
    NO : sonst */
  bool isActionSent();

  /** @return YES oder NO
    YES : wenn neue Sensorinfos ins Weltmodell integriert wurden 
    NO : sonst
    hat den gleichen Wert wie die WeltmodellFunktion*/
  bool isNewInfo();

  /** @return geschaetzte Restzeit bis zum Absenden 
    der Aktion in Millisekunden */
  int remainingTime();

  /// @return aktueller ServerTakt
  int soccerServerSimStep();    

  /// @return String mit Klassenbezeichnung
  string idString() const;

  /// @return Zeit seit Start des Agenten in Millisekunden
  int getAgentTime() const;

  /** versetzt den Agenten-Prozess in den sleep-modus,
      bis controller() == YES 
      ACHTUNG: dieser Schlaf kann recht lange dauern (20ms)
      und der Agent ist "muede", wenn er wieder aufwacht,
      d.h. im Takt des Aufwachens kennen gegebenenfalls
      nicht alle Nebenaktionen gesendet werden. */
  void agentSleep();

  /** Funktion aus dem alten Timer: Liefert Zeiger auf den Effector zurueck. 
      wird von ForceContinuation in Option.cc verwendet */
  Effectors* effector() const;

  Action lastMainAction() const;


 private:
  Sensors* theSensors;
  Effectors* theEffectors;
  Deliberator* theDeliberator;
  Neck* theNeck;
  Say* theSay;
  BodyInfos theLastBodyInfos;
  
   /** Aktueller Zeitabstand in ms zwischen zwei Sichtinformationen */ 
  int viewTimeStep() const;


  /// @returnSetzt die absolute Zeit in s seit 1970 bis zum Start des Agenten
  void setConstructionTime();


  /** absolute Zeit in s seit 1970 bis zum Start des Agenten
      alle anderne Zeiten sind relativ zu dieser Zeit in ms angegeben */
  long timeOfConstruction;

  /** jeweils aktueller Zeitpunkt in ms, bezogen auf timeOfConstruction */
  int now;

  //  Zeitpunkt an dem der Controller zum letzten Mal aufgerufen wurde (Debug)
  int timeOfLastControllerCall;

  /** Zeitpunkt des letzten Aufrufs der Sensoren,
    dient dazu die Sensoren nicht zu oft aufzurufen */
  int timeOfLastSense;

  /** Zeitpunkt des vorletzten Aufrufs der Sensoren,
    dient als untere Abschaetzung fuer den Zeitpunkt des Eintreffens 
    der aktuellen Informationen */
  int timeOfLastButOneSense;
 
 /** Untere Abschaetzung fuer den Zeitpunkt der letzten Sichtinformationen    
   in ms*/
  int timeOfLastSeeInfo;

 /** Abschaetzung fuer den Zeitpunkt des letzten Taktwechsels
    in ms, durch timeOfLastButOneSense, muss keine untere Abschaetzung
    sein, da der Timer nach Taktwechsel mehrfach nach Informationen
    gefragt haben Kann, bevor er zum ersten mal welche erhalten hat

    LSG: bei Debuggen koennte man die Konstante in ms messen,
    um die das SenseBody verzoegert eintrifft */
  int timeOfLastSimStep;
 
  /** letzter Simulationsschritt des Servers
      lt. theSensors->simStepOfSoccerServer()
      in dem der Timer aktiv war*/
  int simStep;
 

  /** Zeitpunkt des letzten Abschickens einer Action in ms */
  int timeOfLastSentAction;

  /** Zeitpunkt des letzten Abschickens einer SayAction in ms */
  int timeOfLastSayAction;

  /** Zeitpunkt des letzten Abschickens von Const-Log Daten in ms */
  int timeOfLastStampFlush;

  /** Zeitpunkt des letzten Abschickens von Var-Log Daten in ms */
  int timeOfLastInfoFlush;

  /** wie gleichnamige Funktion */
  bool theIsNewInfo;

  /** hier wird gespeichert, ob beim controlleraufruf eine entsprechende
    Action gesendet werden soll */
  bool haveToSendMainAction;
  bool haveToSendViewAction;
  bool haveToSendNeckAction;

  /** hier wird gespeichert, ob eine entsprechende Action schon
    im jeweiligen Zeitrahmen gesendet wurde */
  bool didSentMainAction;
  bool didSentViewAction;
  bool didSentNeckAction;
  bool didSentSayAction;

  /** zeigt an, ob bald MainAction gesendet wird, damit
    Nebenaktionen warten */
  bool bewareNextMainAction;


  /** Parameter, der Zeitspanne in ms angibt nach der der Controller bei 
     einem Aufruf wieder aktiv werden soll */
  int controllerTimeStep;
  
  /** Parameter, der deadline im Takt in ms angibt,
    fuer Absenden der Aktion */
  int actionDeadline;

 /** Parameter, der den Zeitabstand in ms angibt,
    zwischen dem Absenden zweier Aktion */
  int timeBetweenTwoActions;

 /** Parameter, der den Zeitabstand in ms angibt,
   zwischen dem Absenden zweier SayAktion */
  int timeBetweenTwoSayActions;

  /** Parameter, der die Mindestzeit angibt die auf einem System
    geschlafen werden kann */
  int systemDependMinSleepTime;

 
  /// @return String der die Kurzkennung der Klasse enthlt
  string theIdString;

  // diese Variable hlt die letzte gesendete Hauptaktion (wichtig fr den Neck)
  Action theLastMainAction;

 
};

//  Globale & Inline Definitionen


/// Globaler Zeiger auf das aktive Timer-Objekt.
extern Timer *globalTimer;


inline string Timer::idString() const
{
   return theIdString;
}


inline int Timer::viewTimeStep() const {
  if ( theLastBodyInfos.viewQuality() == VIEW_QUALITY_HIGH ){ 
    if ( theLastBodyInfos.viewWidth() == VIEW_WIDTH_NARROW ) return ( globalConf->sendStep / 2 );
    else if ( theLastBodyInfos.viewWidth() == VIEW_WIDTH_NORMAL ) return ( globalConf->sendStep );
    else return ( globalConf->sendStep * 2 );
  }
  else
  {
    if ( theLastBodyInfos.viewWidth() == VIEW_WIDTH_NARROW ) return ( globalConf->sendStep / 4 );
    else if ( theLastBodyInfos.viewWidth() == VIEW_WIDTH_NORMAL ) return ( globalConf->sendStep / 2 );
    else return ( globalConf->sendStep );
  }
}
 
inline Effectors* Timer::effector() const
{
  return theEffectors;
}

inline  bool Timer::isEmergency() const
{
 if ( getAgentTime() - timeOfLastSimStep >= actionDeadline + 10 )
   // +10 bedeutet echt spaet
   return YES;
 else
   return NO;
}

inline bool Timer::isMainActionSent()
{
  return didSentMainAction;
}

inline bool Timer::isViewActionSent()
{
  return didSentViewAction;
}

inline bool Timer::isNeckActionSent()
{
  return didSentNeckAction;
}

inline bool Timer::isSayActionSent()
{
  return didSentSayAction;
}

inline bool Timer::isActionSent()
{
  return ( didSentMainAction ||
           didSentViewAction ||
           didSentNeckAction ||
           didSentSayAction     ) ;
}

inline bool Timer::isNewInfo()
{
 return theIsNewInfo;
}


inline int Timer::remainingTime()
{
  return ( timeOfLastSimStep + actionDeadline - getAgentTime() );
}


inline int Timer::soccerServerSimStep()
{
  return simStep;
} 


inline int Timer::getAgentTime() const
{
  static timeval t; 
  gettimeofday( &t, NULL );

  return (t.tv_sec - timeOfConstruction) * 1000 + (t.tv_usec / 1000); 
}

inline void Timer::setConstructionTime()
{
  timeval t; 
  gettimeofday( &t, NULL );
  // gewaehrleistet "normalerweise" "gleiche Startzeiten" der Agenten
  timeOfConstruction = t.tv_sec - (t.tv_sec % 1000);
}

inline Action Timer::lastMainAction() const
{
  return theLastMainAction;
}

#endif
