#ifndef _VECTOR_H_
#define _VECTOR_H_

#include <stream.h>

#include "Ath98String.h"
#include "Common.h"
#include "Exception.h"
#include "Log.h"

/*  Koordinatensysteme, mit denen ein Vektor initialisiert werden
  kann. Moegliche Werte sind {CARTESIAN, POLAR}. */
enum CoordinateSystemEnum {CARTESIAN, POLAR};

/*  Winkelmesssysteme, mit denen ein Winkel gesetzt werden
  kann. Moegliche Werte sind {RADIAN, DEGREE}. */
enum AngleSystemEnum {RADIAN, DEGREE};

/**
  Diese Klasse beschreibt einen 2D-Vektor. 
  Der Vektor akzeptiert Lesen und Schreiben seiner Werte sowohl in
  Polarkoordinaten als auch in kartesischen Koordinaten.

  Per Konvention werden die Werte so interpretiert:

  \begin{description}
  \item[Kartesisches Koordinatensystem:]
  x-Achse nach rechts, y-Achse nach oben. Bei Angabe als Paar: (x, y)

  \item[Polarkoordinatensystem]
  Winkel gegen den Uhrzeigersinn im Bogenmass [0,2*Pi]. Bei Laenge 0 ist auch
  Winkel 0. Bei Angabe als Paar: (Laenge, Winkel)
  \end{description}


  Intern werden momentan nur die kartesischen Koordinaten gefuehrt, die
  Polarkoordinaten jeweils dynamisch ausgerechnet.

  Die Klasse prueft vor Berechnungen <b>nicht</b>, ob die Komponenten gleich 
  UNDEF_DOUBLE sind. Werden zwei undefinierte Vektoren z.B. addiert, so ist das
  Ergebnis nicht unbedingt ein undefinierter Vektor (d.h. er antwortet auf
  isValid() nicht unbedingt mit NO).

  {\small {\bf Quelltext}

  [\URL[Header]{src/Vector.h} | [\URL[Implementierung]{src/Vector.cc}] ]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Vector.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:08 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class Vector
{
public:
  
  /** Interne Fehlerklasse. */
  class VectorException: public Exception {
  public:
    VectorException(const Ath98String& aDescription )
      : Exception( aDescription ) {}
  };
  
  /** Neuer Vektor.
    @param value1 Die x-Koordinate bei kartesischen, die Laenge bei
    Polarkoordinaten.
    @param value2 Die y-Koordinate bei kartesischen, der Winkel im Bogenmass 
    bei Polarkoordinaten.
    @param aCoordinateSystem. Das Koordinatensystem, in dem die beiden vorigen
    Werte zu interpretieren sind. CARTESIAN ist default.
    @param anAngleSystem. Das Winkelmesssystem, in dem der Winkel bei
    einer Polarkoordinatenangabe angegeben ist. RADIAN ist default.
    @param ist einer der Werte undefiniert wird der Vektor auf (UNDEF_DOUBLE,UNDEF_DOUBLE) gesetzt
    @exception VectorException falls: unbekanntes Koordinatensystem, unbekannte  Winkelart. */
  Vector(
	 double value1, 
	 double value2, 
	 CoordinateSystemEnum aCoordinateSystem = CARTESIAN,
	 AngleSystemEnum anAngleSystem = RADIAN );
  
  /** Neuer undefinierter Vektor. */
  Vector();

  /** Gibt die x-Koordinate zurueck. */
  double x() const;

  /** Gibt die y-Koordinate zurueck. */
  double y() const;
  
  /** Setzt die x- und die y-Koordinate. */
  void setCartesian(const double anXValue, const double anYValue);
  
  /** Gibt die Laenge zurueck.(falls der Vektor ungueltig ist UNDEF_DOUBLE) */
  double length() const;

  /** Gibt Quadrat der Lnge zurck, dies ist wesentlich schneller, da eine
      kostenspielige Wurzeloperation fehlt, vor allem bei innerster
      Schleife und Vergleichen diese benutzen ! */
  double length_sqr() const;
  
  /** Gibt den Winkel zurueck.(falls der Vektor ungueltig ist UNDEF_DOUBLE)
    @exception VectorException falls: unbekannte  Winkelart. */
  double angle( AngleSystemEnum anAngleSystem = RADIAN ) const;
  
  /** Setzt die Laenge und den Winkel des Vektors.
      Falls Laenge oder Winkel undefiniert ist, wird der Vektor auf (UNDEF_DOUBLE,UNDEF_DOUBLE) gesetzt
    @exception VectorException falls: unbekannte  Winkelart. */
  void setPolar( double aLength, double anAngle,
		 AngleSystemEnum anAngleSystem = RADIAN );

  /** Setzt Werte neu.
    @param value1 Die x-Koordinate bei kartesischen, die Laenge bei
    Polarkoordinaten.
    @param value2 Die y-Koordinate bei kartesischen, der Winkel bei
    Polarkoordinaten.
    @param aCoordinateSystem Das Koordinatensystem, in dem die beiden vorigen
    Werte zu interpretieren sind. CARTESIAN ist default.
    @param anAngleSystem. Das Winkelmesssystem, in dem der Winkel bei
    einer Polarkoordinatenangabe angegeben ist. RADIAN ist default.
    @exception VectorException falls: unbekanntes Koordinatensystem, unbekannte  Winkelart. */
  void set( 		
	   double value1, 
	   double value2, 
	   CoordinateSystemEnum aCoordinateSystem = CARTESIAN,
	   AngleSystemEnum anAngleSystem = RADIAN  );

  /* Hack fr P */
  void setX( double x );
  void setY( double y);
  /* Hack Ende */

  /** Prueft, ob der Vektor komplett definiert ist.. 
      @return YES falls beide Werte(x,y bzw. Laenge,Winkel) definiert sind, sonst NO */
  bool isValid() const;

  /** Setzt die Laenge auf 1. Der Winkel bleibt gleich */
  void normalize();

  /** Gibt einen orthogonalen Vektor zur\"uck. Es wird gegen
      den Uhrzeigersinn gedreht. */
  Vector orthogonal();
    
  /** Vektoraddition.  */
  Vector operator + (const Vector& v2) const;
  
  /** Vektorsubtraktion. */
  Vector operator - (const Vector& v2) const;
  
  /** die skalare Multiplikation von links */
  friend Vector operator * (double aScalarValue, const Vector& aVector);

  /** skalare Multiplikation zweier Vektoren */
  const double operator * (const Vector& v2);

  /** TM Multiplikation mit double komponentenweise */
  Vector operator * (double aScalar) const;

  /** TM Division mit double komponentenweise */
  Vector operator / (double aScalar) const;

    /** Addition des Vektor v2 zum aktuellen Wert. */
  const Vector& operator += (const Vector& v2);

  /** Subtraktion des Vektors v2 vom aktuellen Wert. */
  const Vector& operator -= (const Vector& v2);

  /** Vergleich des Vectors v2 mit dem aktuellen Vektor.
    Es wird \Ref{isAlmostEqual} verwendet.
    @return YES, wenn die beiden Vektoren praktisch identisch sind, sonst NO */
  bool operator == ( const Vector v2 ) const;

  /** Vergleich des Vectors v2 mit dem aktuellen Vektor auf Ungleichheit.
    Es wird \Ref{isAlmostEqual} verwendet.
    @return NO, wenn die beiden Vektoren praktisch identisch sind, sonst YES */
  bool operator != ( const Vector v2 ) const;

  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const Vector& );

  /** kompakte Ausgabe auf einen Stream fr constant Log */
    void toStampStream( ostream& ) const;

  /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal.
      Es werden kartesische UND Polarkoordinate ausgegeben.
      Die Ausgabe der Polarkoordinaten erfolgt in Grad.
      Ausserdem wird auf ungueltige Vektoren hingewiesen. */
  Ath98String toString() const;
  
private:
  /** Die x-Komponente. */
  double theX;
  /** Die y-Komponente. */
  double theY;
};


inline double Vector::x() const
{
  return theX;
}

inline double Vector::y() const
{
  return theY;
}

inline void Vector::setCartesian(const double anXValue, const double anYValue)
{
  theX=anXValue;
  theY=anYValue;
}

inline void Vector::setX( double x )
{
  theX = x;
}

inline void Vector::setY( double y )
{
  theY = y;
}
/* ENDE HACK */

inline bool Vector::isValid() const
{
  return !(theX == UNDEF_DOUBLE || theY == UNDEF_DOUBLE);
}

inline Vector operator * (double aScalarValue, const Vector& aVector) 
{
  return Vector(aScalarValue * aVector.x(), aScalarValue * aVector.y() );
}

inline const double Vector::operator * (const Vector &v2 )
{
  return (x()*v2.x() + y()*v2.y());
}


inline Vector Vector::operator+ (const Vector& v2) const
{
  return Vector(x()+v2.x(),y()+v2.y());
}



inline Vector Vector::operator- (const Vector& v2) const
{
  return Vector(x()-v2.x(),y()-v2.y());
}

inline const Vector& Vector::operator+= (const Vector& v2)
{
  theX += v2.x();
  theY += v2.y();
  return *this;
}

inline const Vector& Vector::operator-= (const Vector& v2)
{
  theX -= v2.x();
  theY -= v2.y();
  return *this;
}

inline bool Vector::operator == ( const Vector v2 ) const
{
  return isAlmostEqual( x(), v2.x() ) && isAlmostEqual( y(), v2.y() );
}
    
inline bool Vector::operator != ( const Vector v2 ) const
{
  return !operator==( v2 );
}

inline double Vector::length() const 
{   
  if ( theX == UNDEF_DOUBLE || theX == UNDEF_DOUBLE ) { 
    return UNDEF_DOUBLE;  
  } 
  return sqrt(theX*theX+theY*theY);
} 
   
inline double Vector::length_sqr() const
{
  if ( theX == UNDEF_DOUBLE || theX == UNDEF_DOUBLE ) { 
    return UNDEF_DOUBLE;  
  } 

  return theX*theX+theY*theY;
}  

inline ostream& operator<< (ostream& os, const Vector& v1 )
{
  os << " kartesisch : (" << v1.x()      << "," << v1.y() <<  
            ") polar : (" << v1.length() << "," << v1.angle(DEGREE) <<  
    " ) "  ;
  return os;
}

inline void Vector::toStampStream( ostream& os) const {
  os << 
    " " << x() << 
    " " << y();
}



inline int orientation(const Vector& a, const Vector& b, const Vector& c)
{ double d1 = (a.x() - b.x()) * (a.y() - c.y());
  double d2 = (a.y() - b.y()) * (a.x() - c.x()); 
  if (d1 == d2) return 0; else return (d1 > d2) ? +1 : -1;
}

#endif

 