
/*
 * Aktueller Stand
 *
 *   $RCSfile: Coach.cc,v $
 *   $Revision: 1.13 $
 *   $Date: 2001/08/02 11:46:53 $
 *   $Author: wendler $ 
 *
 */
 
#include "./Coach.h"
#include "./Deliberator.h"
#include "./Timer.h"
#include "./TeamInfo.h"
#include "../basic/CommonCoach.h"
#include "../serverInterface/SensorsServer.h"
#include "../serverInterface/EffectorsServer.h"
#include "../serverInterface/SensorsRecord.h"
#include "../serverInterface/EffectorsRecord.h"
#include "../../serverInterface/Communicator.h"
#include "../../worldmodel/WorldModelInterface.h"
#include "../../basic/Log.h"
#include "../../basic/Configuration.h"
#include "../../worldmodel/Situation.h"

int main(int argc, char* argv[])
{
  Coach* theCoach;

  try {
    theCoach = new Coach( argc, argv );
    theCoach->run();
  }
  catch( Exception e ) {
    cerr << "Error: " << e.toString() << endl;
    delete theCoach;
    ERROR( e.toString() );
    globalLog->flushInfoStream();
    globalLog->flushStampStream();
    exit( -1 );
  }
  return 0;
}


Coach::Coach(int argCount, char* argv [])
{
  teamName = "AT_Humboldt";
  serverName = "localhost";
  serverPort = 6002;
  serverConfigurationFile = "server.conf";
  agentConfigurationFile = "agent.conf";
  coachConfigurationFile = "coach.conf";
  recFile = "";
  extendedConfigurationFiles = vector<Ath98String>();
  isHelpWanted = NO;
  useInteractive = NO;
  globalLog = new Log( 1 );

  readParameter( argCount, argv );
  if( isHelpWanted ) {
    showHelp( Ath98String(argv[0]) );
    exit(0);
  }

  // Logfile anlegen
  if( !globalLog->isFileReady() ) {
    globalLog->createFiles( COACH_NUM, teamName );
    INFO( Log::ANY, 0, idString(), globalLog->infoStream <<
          "Logfiles created.\n" );
  }

  globalConf = new Configuration();
  globalConf->coachUseInteractive = useInteractive;

  globalConf->readConfigurationFile( serverConfigurationFile );
  globalConf->readConfigurationFile( agentConfigurationFile );
  globalConf->readConfigurationFile( coachConfigurationFile );
  if( extendedConfigurationFiles.size() != 0 ) {
    for( unsigned int i = 0;  i < extendedConfigurationFiles.size();  i++ ) {
      globalConf->readConfigurationFile( extendedConfigurationFiles[i] );
    }
  }
  globalLog->init( globalConf );
  globalConf->logCurrentParameterValues();

  serverPort = globalConf->coachPortNumber;
  startComponents();
}

Coach::~Coach () 
{
  delete theDeliberator;
}

void Coach::startComponents() {
  initRandomNumberGenerator();

  // heterogene Spieler
  globalPlayerTypes = new PlayerTypes;
  MessageCoach::setTeamname( teamName );

  // Sensoren und Effektoren
  if( recFile == "" ) {
    theCommunicator = new Communicator( serverName, serverPort );
    theSensors = new SensorsServer( theCommunicator );
    theEffectors = new EffectorsServer( theCommunicator );
  }
  else {
    cout << "Using recorded file " << recFile << " as input." << endl;
    theSensors = new SensorsRecord( recFile );
    theEffectors = new EffectorsRecord();
    globalConf->coachWithRecordFile = YES;;
  }

  // globales Weltmodell, Planen und Ueberlegen
  globalWorldModel = new WorldModelInterface( theSensors );
  globalTeamInfo = new TeamInfo();
  
  theDeliberator = new Deliberator( theSensors, theEffectors );
  globalTimer = new Timer( theSensors, theEffectors, theDeliberator );

  theEffectors->enterField( teamName );
  globalWorldModel->setTeamName( teamName );
  globalWorldModel->setSide( theSensors->checkEnterFieldForSide() );
  INFO( Log::ANY, 2, idString(), globalLog->infoStream <<
	"\nEntered field for team: " << globalWorldModel->teamName() <<
        " at side: " << (globalWorldModel->side() == LEFT_SIDE_ID ?
                         "LEFT" : "RIGHT") << "\n" );
  theSensors->checkServerParams();
  theEffectors->openEyes();
  INFO( Log::ANY, 2, idString(), globalLog->infoStream <<
        "Eyes opened.\n" );

  theDeliberator->initialize();
}

void Coach::run() {
  cout << "Coach starts deliberating." << endl;
  INFO( Log::ANY, 2, idString(), globalLog->infoStream <<
        "Coach starts deliberating.\n" );
  globalLog->flushInfoStream();
  theDeliberator->deliberate();
}


void Coach::readParameter( int argCount, char* argv [] ) {
  int argIndex;
  Ath98String parameter;

  for( argIndex = 1;  argIndex < argCount;  argIndex++ ) {
    parameter = Ath98String( argv[argIndex] );  
    if( parameter == "-help" || parameter == "-h" ) { 
      isHelpWanted = YES;
      return;  
    }
    else if( parameter == "-interactive" || parameter == "-i" ) { 
      useInteractive = YES;
      return;  
    } 
    else {  // zweistellige Parameter 
      argIndex++;  
      if( !(argIndex < argCount) ) {
	throwException(InitFailedException( "Missing parameter after option " +
            Ath98String( argv[argIndex-1] ) + ". Aborting initialization."));  
      }
      if( parameter == "-team" ) { 
	teamName = Ath98String( argv[argIndex] );  
      } 
      else if( parameter == "-host" ) { 
	serverName = Ath98String( argv[argIndex] );  
      } 
      else if( parameter == "-port" ) { 
	serverPort = atoi( argv[argIndex] );
      } 
      else if( parameter == "-sfile" ) {
	serverConfigurationFile = Ath98String( argv[argIndex] );
      } 
      else if( parameter == "-afile" ) { 
	agentConfigurationFile = Ath98String( argv[argIndex] );
      } 
      else if( parameter == "-cfile" ) { 
	coachConfigurationFile = Ath98String( argv[argIndex] );
      } 
      else if( parameter == "-efile" ) { 
	extendedConfigurationFiles.push_back( Ath98String( argv[argIndex] ) );
      } 
      else if( parameter == "-rec" ) {
        recFile = Ath98String( argv[argIndex] );
      }
      else { 
	argIndex--;
        cerr << "Unknown option in Commandline: " << 
	  Ath98String( argv[argIndex] ) << " is ignored." << endl;  
      }
    } 
  }  // Ende for-Schleife 
} 

void Coach::initRandomNumberGenerator() 
{
  static long probabilityState[32] = { 
    3, 
    0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342, 
    0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd, 
    0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86, 
    0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7, 
    0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc, 
    0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
    0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b, 
    0xf5ad9d0e, 0x8999220b, 0x27fb47b9 
  };  
  unsigned int seed = 1;  
  int n = 128;  
   
  initstate(seed, (char *)probabilityState, n);  
  setstate( (char *)probabilityState );
} 

void Coach::showHelp( Ath98String programName ) {
  cout << endl << "*******  AT Humboldt 2001 - Coach  *******" << endl << endl;
  cout << "Humboldt University Berlin" << endl;
  cout << "Department of Computer Science" << endl;
  cout << "AI Group" << endl;
  cout << endl << endl;
  cout << "Usage: " << programName << " [options]" << endl;
  cout << "Options:" << endl;

  cout << "  -h, -help          " << "\t" <<
    "this help info" << endl;
  cout << "  -i, -interactive   " << "\t" <<
    "use the coach interactive (per stdin)" << endl;
  cout << "  -host HOSTNAME     " << "\t" <<
    "the host where the server runs (localhost)" << endl;
  cout << "  -team TEAMNAME     " << "\t" <<
    "the name of the team (AT_Humboldt)" << endl;
  cout << "  -port PORT_NUMBER  " << "\t" <<
    "the port at which the server waits for the coach (6002)" << endl;
  cout << "  -sfile SERVER_FILE " << "\t" <<
    "the configuration file of the soccerserver (server.conf)" << endl;
  cout << "  -afile AGENT_FILE  " << "\t" <<
    "the configuration file of the agent (agent.conf)" << endl;
  cout << "  -cfile COACH_FILE  " << "\t" <<
    "the configuration file of the coach (coach.conf)" << endl;
  cout << "  -efile FILE  " << "\t" <<
    "an extended configuration file of the agent or coach (extended.conf) " <<
    "- multiple use allowed" << endl;
  cout << "  -rec REC_FILE  " << "\t" <<
    "a recorded game which is used as input for the coach" << endl;
}



