
/*
 * Aktueller Stand
 *
 *   $RCSfile: WorldModelCore.cc,v $
 *   $Revision: 1.10 $
 *   $Date: 2001/07/19 16:09:07 $
 *   $Author: wendler $ 
 *
 */

#include "./WorldModelCore.h"
#include "../basic/CommonCoach.h"
#include "../basic/Timer.h"
#include "../serverInterface/Sensors.h"

#include "../../worldmodel/WorldModelInterface.h"
#include "../../basic/InvalidIdException.h"

// Konstruktor/Destruktor
WorldModelCore::WorldModelCore( Sensors *sensors) :
  theSensors( sensors )
{ 
  theSituations = vector<Situation>( 0 );
  theSituations.push_back( Situation(0) );
  theCurrentSituationIndex = 0;
  theTeamName = Ath98String( "Unknown Team" );
  thePlayerNumber = COACH_NUM;
  theSide = NO_SIDE_ID;
  theScore[0] = 0; theScore[1] = 0;
  theInitiative = NO_TEAM_ID;
  theGameState = WorldModelInterface::BEFORE_KICK_OFF;
  theIdString = "WorldModelCore";
}

WorldModelCore::~WorldModelCore()
{  
}


// Anfragen und Setzen von Coach-Werten
int WorldModelCore::playerNumber() const 
{
  return thePlayerNumber;
}

const Ath98String& WorldModelCore::teamName() const
{
  return theTeamName;
}

int WorldModelCore::side() const 
{
  return theSide;
} 
 
int WorldModelCore::id() const
{
  return COACH_ID;
}

void WorldModelCore::setPlayerNumber( int newPlayerNumber, bool isGoalie ) 
{ 
  thePlayerNumber = newPlayerNumber;
}

void WorldModelCore::setTeamName( const Ath98String& newTeamName )
{
  theTeamName = newTeamName;
}

void WorldModelCore::setSide( int newSide ) 
{
  theSide = newSide;
}

// Anfragen zur aktuellen oder aelteren Situationen
const Situation& WorldModelCore::situation( int situationOffset = 0 ) const 
{ 
  int index = theCurrentSituationIndex - situationOffset;
  if( index < 0 ) {
    throwException( "TimeStep offset too large." );
  }
  else {
    return theSituations[index];
  }
} 


const ModelledPlayer& WorldModelCore::player( int playerId ) const 
{ 
  return theSituations[theCurrentSituationIndex].playerWithId( playerId ); 
} 
 
const ModelledBall& WorldModelCore::ball() const 
{ 
  return theSituations[theCurrentSituationIndex].ball();
} 

int WorldModelCore::simStep() const 
{
  return theSituations[theCurrentSituationIndex].simStep();
}

int WorldModelCore::gameState() const
{
  return theGameState;
}

int WorldModelCore::initiative() const
{
  return theInitiative;
}

bool WorldModelCore::hasInitiative( int aTeamId ) const 
{ 
  bool answer;
  switch ( aTeamId ) {
  case OUR_TEAM_ID:
    answer = ( initiative() == OUR_TEAM_ID ||
               initiative() == BOTH_TEAMS_ID );
    break;
  case THEIR_TEAM_ID:
    answer = ( initiative() == THEIR_TEAM_ID ||
               initiative() == BOTH_TEAMS_ID );
    break;
  default:
    throwException( InvalidIdException(
        "Only OUR_TEAM_ID and THEIR_TEAM_ID are allowed as parameters." ) );
  }
  return answer;
} 

bool WorldModelCore::isTimeElapsingGameState() const
{
  return( gameState() != WorldModelInterface::BEFORE_KICK_OFF &&
          gameState() != WorldModelInterface::TIME_OVER &&
          gameState() != WorldModelInterface::GOAL &&
          gameState() != WorldModelInterface::OFFSIDE );
}

void WorldModelCore::setSimStep( int newValue )
{
  theSituations[theCurrentSituationIndex].setSimStep( newValue );
}

void WorldModelCore::setGameState( int newGameState ) 
{ 
  theGameState = newGameState;
} 

void WorldModelCore::setInitiative( int newInitiative )  
{ 
  if ( newInitiative == OUR_TEAM_ID || newInitiative == THEIR_TEAM_ID || 
       newInitiative == BOTH_TEAMS_ID ) { 
    theInitiative = newInitiative;  
  } 
  else if ( newInitiative == LEFT_SIDE_ID || newInitiative == RIGHT_SIDE_ID) { 
    if ( side() == newInitiative ) { 
      theInitiative = OUR_TEAM_ID;   
    } 
    else { 
      theInitiative = THEIR_TEAM_ID;  
    } 
  } 
  else { 
    theInitiative = NO_TEAM_ID;  
  }
}


// Allgemeine Anfragen und Setzmoeglichkeiten

int WorldModelCore::score( int aTeamId ) const
{ 
  switch ( aTeamId ) {
  case OUR_TEAM_ID: 
    return theScore[0];  
    break;  
  case THEIR_TEAM_ID: 
    return theScore[1];  
    break;
  case LEFT_SIDE_ID:
    return ( ( side() == LEFT_SIDE_ID ) ? theScore[0] : theScore[1] );
    break;
  case RIGHT_SIDE_ID:
    return ( ( side() == RIGHT_SIDE_ID ) ? theScore[0] : theScore[1] );
    break;
  default: 
    throwException( InvalidIdException(  
        "Score for unknown team id demanded." ) );  
  }
}

void WorldModelCore::setScore(int newScore, int aTeamId )
{ 
  switch ( aTeamId ) { 
  case OUR_TEAM_ID: 
    theScore[0] = newScore;  
    break;  
  case THEIR_TEAM_ID: 
    theScore[1] = newScore;  
    break;  
  case LEFT_SIDE_ID:
    if( side() == LEFT_SIDE_ID ) {
      theScore[0] = newScore;
    }
    else {
      theScore[1] = newScore;
    }
    break;
  case RIGHT_SIDE_ID:
    if( side() == RIGHT_SIDE_ID ) {
      theScore[0] = newScore;
    }
    else {
      theScore[1] = newScore;
    }
    break;
  default: 
    throwException( InvalidIdException(  
        "You try to set score for an unknown team id." ) );  
  }
}


const Field& WorldModelCore::field() const
{
  return theField;
}


// Vorzugsrichtungen
Vector WorldModelCore::preferredKickDirection( const Vector& pos ) const
{ 
  INFO( Log::WORLD_MODEL, 6, idString(), globalLog->infoStream <<
        "    ENTER WorldModel::preferredKickDirection\n" );

  Vector toTarget;
  const Vector WING_ATTRACTOR( 0, 34 );
  Vector opponentAttractor;
  if ( pos.y() > 0 ) {
    opponentAttractor.set( 52.5, 4 );
  }
  else {
    opponentAttractor.set( 52.5, -4 );
  }

  const double maximumSpeed = globalConf->ballSpeedMax;
  const double slowSpeed = maximumSpeed / 5;
  // Die Vorzugsrichtung ist abhaengig von der Region.
  RegionOnField regionOfPos = region( pos );
  INFO( Log::WORLD_MODEL, 8, idString(), globalLog->infoStream <<
	"I am in region " << (int)regionOfPos << "\n" );
  switch ( regionOfPos ) {
  case TO_UPPER:
    // Raus auf den oberen Fluegel
    toTarget = WING_ATTRACTOR - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );
    break;  
  case TO_LOWER: 
    // Raus auf den unteren Fluegel 
    toTarget = ( -1.0 * WING_ATTRACTOR ) - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );  
    break;  
  case REST: 
  case UPPER_MIDDLE: 
  case LOWER_MIDDLE: 
    // waagrecht rueber in die gegnerische Haelfte 
    toTarget = opponentAttractor;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );  
    break;  
  case UPPER: 
  case LOWER: 
    // weg vom Rand, direkt auf das Tor des Gegners
    toTarget = opponentAttractor - pos;  
    toTarget.setPolar( slowSpeed, toTarget.angle() );  
    break;  
  case ENEMY_GOAL: 
    // direkt aufs gegnerische Tor 
    toTarget = opponentAttractor - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );  
    break;  
  case UPPER_EDGE: 
  case LOWER_EDGE: 
    toTarget = Vector( opponentAttractor.x() - 11, opponentAttractor.y() ) - 
      pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );  
    break;  
  default: 
    throwException( InvalidIdException( "Invalid region number." ) );  
  } 
  INFO( Log::WORLD_MODEL, 6, idString(), globalLog->infoStream <<
	"Preferred kick direction is " << toTarget << "\n" );

  return toTarget;  
} 

Vector WorldModelCore::backKickDirection( const Vector& pos ) const 
{ 
  INFO( Log::WORLD_MODEL, 7, idString(), globalLog->infoStream <<
        "    ENTER WorldModel::backKickDirection\n" );
  
  Vector toTarget;
  const Vector WING_ATTRACTOR( 15, 34 );  
  Vector opponentAttractor;  
  if ( pos.y() > 0 ) { 
    opponentAttractor.set( 52.5, 4 );  
  } 
  else { 
    opponentAttractor.set( 52.5, -4 );  
  } 
  Vector selfAttractor;  
  if ( pos.y() > 0 ) { 
    opponentAttractor.set( -52.5, 30 );  
  }
  else {
    opponentAttractor.set( -52.5, -30 );
  }

  const double maximumSpeed = globalConf->ballSpeedMax / 1.5;
  const double slowSpeed = maximumSpeed / 4;  
  // Die Vorzugsrichtung ist abhaengig von der Region. 
  RegionOnField regionOfPos = region( pos );  
  INFO( Log::WORLD_MODEL, 8, idString(), globalLog->infoStream <<
	"I am in region " << (int)regionOfPos << "\n" );
  switch ( regionOfPos ) { 
  case TO_UPPER: 
    // In die Mitte 
    toTarget = WING_ATTRACTOR - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() + Ath98_PI );  
    break;  
  case TO_LOWER: 
    // In die Mitte 
    toTarget = ( -1.0 * WING_ATTRACTOR ) - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() + Ath98_PI );  
    break;  
  case REST: 
  case UPPER_MIDDLE:
  case LOWER_MIDDLE: 
    // Gerade zurueck 
    toTarget = opponentAttractor;  
    if( pos.x() < 40 ){ 
      toTarget.setPolar( slowSpeed, toTarget.angle() + Ath98_PI );  
    } 
    else{ 
      toTarget.setPolar( maximumSpeed, toTarget.angle() + Ath98_PI );  
    } 
    break;  
  case UPPER: 
  case LOWER: 
    // Auf der Linie zurueck 
    toTarget = selfAttractor - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );  
    break;  
  case ENEMY_GOAL: 
    // Auf die Fluegel zurueck 
    toTarget = opponentAttractor - pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() + Ath98_PI );  
    break;  
  case UPPER_EDGE: 
  case LOWER_EDGE:
    toTarget = Vector( opponentAttractor.x() - 11, opponentAttractor.y() ) - 
      pos;  
    toTarget.setPolar( maximumSpeed, toTarget.angle() );  
    break;  
  default: 
    throwException( InvalidIdException( "Invalid region number." ) );  
  } 
  INFO( Log::WORLD_MODEL, 7, idString(), globalLog->infoStream <<
        "Preferred back-kick direction is " << toTarget << "\n" );
  
  return toTarget;
}

const double WorldModelCore::OUTER_LIMIT = 28.0;
const double WorldModelCore::INNER_LIMIT = 20.0;

WorldModelCore::RegionOnField WorldModelCore::region( const Vector& pos ) const
{
  const double MAIN_DIVIDER = 15;  
  const double WING_DISTRIBUTOR_DIVIDER = -20;  
  const double WING_DISTRIBUTOR_INNER_LIMIT = 7.1;  
  double x = pos.x();  
  double y = pos.y();  
  if ( x < MAIN_DIVIDER ) { 
    // Teil des Spielfelds, der die Mittellinie und das eigene Tor 
    // enthaelt 
    if ( y > OUTER_LIMIT ) { 
      // oben, Feldrand 
      return UPPER;  
    } 
    if ( y > INNER_LIMIT ) { 
      // oben, Streifen darunter 
      return UPPER_MIDDLE;  
    } 
    if ( y < -OUTER_LIMIT ) { 
      // unten, Feldrand 
      return LOWER;  
    } 
    if ( y < -INNER_LIMIT ) { 
      // unten, darueber 
      return LOWER_MIDDLE;
    } 
    if ( x < WING_DISTRIBUTOR_DIVIDER ) { 
      if ( y > WING_DISTRIBUTOR_INNER_LIMIT ) { 
        // Verteiler auf den oberen Fluegel 
        return TO_UPPER;  
      } 
      if ( y < -WING_DISTRIBUTOR_INNER_LIMIT ) { 
        // Verteiler auf den unteren Fluegel 
        return TO_LOWER;  
      } 
    } 
    // Der Rest ist eine T-foermige Region mit dem Querbalken quer zur 
    //Mittellinie und dem Laengsbalken zum eigenen Tor hin 
    return REST;  
  } 
  else { 
    // Teil des Spielfelds, das das gegnerische Tor enthaelt 
    double m;
    double c;  
    // Steigung(m) und Achsenabschnitt(c) einer Geraden berechnen, die 
    // vom oberen Torpfosten zum Punkt (18.5, 34) geht 
    m = ( theField.goalWidth()/2 - theField.fieldWidth()/2 ) /  
      ( theField.fieldLength()/2 - 18.5 );  
    c = theField.fieldWidth()/2 - 18.5 * m;  
    if ( y > m * x + c || y > OUTER_LIMIT ) { 
      // y liegt oberhalb der Geraden oder oberhalb von OUTER_LIMIT 
      return UPPER_EDGE;  
    } 
    // Steigung(m) und Achsenabschnitt(c) einer Geraden berechnen, die 
    // vom unteren Torpfosten zum Punkt (18.5, -34) geht 
    m = ( -theField.goalWidth()/2 + theField.fieldWidth()/2 ) /  
      ( theField.fieldLength()/2 - 18.5 );  
    c = -theField.fieldWidth()/2 - 18.5 * m;  
    if ( y < m * x + c || y < -OUTER_LIMIT ) { 
      // y liegt unterhalb der Geraden oder unterhalb von OUTER_LIMIT 
      return LOWER_EDGE;  
    } 
    // Das uebrigbleibende Rechteck mit zwei abgeschnittenen Ecken. Die kurze 
    // Seite ist das gegnerische Tor 
    return ENEMY_GOAL;  
  }
}


//Einarbeitung und Kommunikation
void WorldModelCore::integrateSensorInfo() 
{ 
  if( theSensors->isNewSeeInfo() ) {
    theSituations[0] = theSensors->situation();
  }
  if( theSensors->isNewRefereeInfo() ) {
    processRefereeInfo();
  }
  if( theSensors->isNewHearInfo() ) {
    processHearInfo();
  }
  if( theSensors->isNewChangePlayerTypeInfo() ) {
    processChangePlayerTypeInfo();
  }
}

void WorldModelCore::processRefereeInfo() 
{
  const vector<MessageCoach>& messages = theSensors->heardMessages();
  for( unsigned int i = 0; i < messages.size(); ++i ) {
    if( messages[i].senderInfo() == REFEREE_ID ) {
      processRefereeMessage( messages[i] );
    }
  }
}

void WorldModelCore::processRefereeMessage( const MessageCoach& message )
{
  const Ath98String& messageString = message.message();
  
  if ( messageString == "before_kick_off" ) { 
    setGameState( WorldModelInterface::BEFORE_KICK_OFF );
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString == "time_over" ) {
    setGameState( WorldModelInterface::TIME_OVER );
    setInitiative(  NO_TEAM_ID );
  }  
  else if ( messageString == "play_on" ) { 
    setGameState( WorldModelInterface::PLAY_ON );
    setInitiative(  BOTH_TEAMS_ID );
  }
  else if ( messageString == "kick_off_l" ) {
    setGameState( WorldModelInterface::KICK_OFF );
    setInitiative(  LEFT_SIDE_ID );
  } 
  else if ( messageString == "kick_off_r" ) {
    setGameState( WorldModelInterface::KICK_OFF );
    setInitiative(  RIGHT_SIDE_ID );
  }  
  else if ( messageString == "kick_in_l" ) { 
    setGameState( WorldModelInterface::KICK_IN );
    setInitiative( LEFT_SIDE_ID );
  } 
  else if ( messageString == "kick_in_r" ) { 
    setGameState( WorldModelInterface::KICK_IN );
    setInitiative(  RIGHT_SIDE_ID );
  } 
  else if ( messageString == "free_kick_l" ) { 
    setGameState( WorldModelInterface::FREE_KICK );
    setInitiative(  LEFT_SIDE_ID );
  }
  else if ( messageString == "free_kick_r" ) {
    setGameState( WorldModelInterface::FREE_KICK );
    setInitiative(  RIGHT_SIDE_ID );
  }
  else if ( messageString == "drop_ball" ) {
    setGameState( WorldModelInterface::DROP_BALL );
    setInitiative(  BOTH_TEAMS_ID );
  }
  else if ( messageString == "corner_kick_l" ) {
    setGameState( WorldModelInterface::CORNER_KICK );
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString == "corner_kick_r" ) {
    setGameState( WorldModelInterface::CORNER_KICK );
    setInitiative( RIGHT_SIDE_ID );
  }
  else if ( messageString == "goal_kick_l" ) {
    setGameState( WorldModelInterface::GOAL_KICK );
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString == "goal_kick_r" ) {
    setGameState( WorldModelInterface::GOAL_KICK );
    setInitiative( RIGHT_SIDE_ID );
  }
  else if ( messageString == "offside_l" ) {
    setGameState( WorldModelInterface::OFFSIDE );
    setInitiative( RIGHT_SIDE_ID );
  }
  else if ( messageString == "offside_r" ) {
    setGameState( WorldModelInterface::OFFSIDE );
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString == "goalie_catch_ball_l" ) {
    setGameState( WorldModelInterface::CATCH_BALL );
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString == "goalie_catch_ball_r" ) {
    setGameState( WorldModelInterface::CATCH_BALL );
    setInitiative( RIGHT_SIDE_ID );
  }
  else if ( messageString.beginsWith( "goal_l" ) ) {
    int newScore = messageString.subString( 7 ).asInt();
    setGameState( WorldModelInterface::GOAL );
    if ( side() == LEFT_SIDE_ID ) {
      setScore( newScore, OUR_TEAM_ID );
    }
    else {
      setScore( newScore, THEIR_TEAM_ID );
    }
    setInitiative( RIGHT_SIDE_ID );
  }
  else if ( messageString.beginsWith( "goal_r" ) ) {
    int newScore = messageString.subString( 7 ).asInt();
    setGameState( WorldModelInterface::GOAL );
    if ( side() == RIGHT_SIDE_ID ) {
      setScore( newScore, OUR_TEAM_ID );
    }
    else {
      setScore( score( THEIR_TEAM_ID ) + 1,THEIR_TEAM_ID );
    }
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString.beginsWith( "foul_l" ) ) {
    setGameState( WorldModelInterface::FOUL );
    setInitiative( RIGHT_SIDE_ID );
  }
  else if ( messageString.beginsWith( "foul_r" ) ) {
    setGameState( WorldModelInterface::FOUL );
    setInitiative( LEFT_SIDE_ID );
  }
  else if ( messageString == "half_time" ) {
    setGameState( WorldModelInterface::HALF_TIME );
    setInitiative( NO_TEAM_ID );
  }
  else if ( messageString == "time_up" ) {
    setGameState( WorldModelInterface::TIME_UP );
    setInitiative( NO_TEAM_ID );
  }
  else if ( messageString == "time_up_without_a_team" ) {
    setGameState( WorldModelInterface::TIME_UP_WITHOUT_A_TEAM );
    setInitiative( NO_TEAM_ID );
  }
  else if ( messageString == "time_extended" ) {
    setGameState( WorldModelInterface::TIME_EXTENDED );
    setInitiative( NO_TEAM_ID );
  }
  else if ( messageString.beginsWith( "goalie_catch_ball" ) ) {
    // Nichts machen. Gleich darauf kommt ein Freekick. Dann wird
    // die Initiative richtig gesetzt.
  }
  else if ( messageString.beginsWith( "freeze" ) ) {
    setGameState( WorldModelInterface::FREEZE );
  }
  else {
    INFO( Log::WORLD_MODEL, 0, idString(), globalLog->infoStream <<
          "ERROR: Unknown referee decision: " << messageString << "\n" );
  }
}

void WorldModelCore::processHearInfo() 
{

}

void WorldModelCore::processChangePlayerTypeInfo()
{

}


Ath98String WorldModelCore::saySituation() 
{
  // to implement
  return Ath98String();
}


// Debugging
string WorldModelCore::idString() const {
  return theIdString;
}

ostream& operator<< (ostream& os, const WorldModelCore& w) {
  os <<
    "WorldModel Info: simStep: " << w.simStep() <<
    "\n  gameState: " << w.gameState() <<
    "\t  initiative: " << w.initiative()  <<
    "\t  score (our-their): " << w.score(OUR_TEAM_ID) << 
    ":" << w.score(THEIR_TEAM_ID) <<
    "\n  current Situation: " <<
    "\n    ball: p(" << w.situation().ball().pos().x() <<
    ", " << w.situation().ball().pos().y() << 
    ") s(" << w.situation().ball().speed().x() <<
    ", " << w.situation().ball().speed().y() << ")";
  os << " sRel(" << w.situation().ball().speed().length() <<
    ", " << degFromRadiant(w.situation().ball().speed().angle()) << ")";
  int n = w.situation().playerCount();
  for( int i = 0; i < n; i++ ) {
    os << "\n    p ";
    if( w.situation().playerWithIndex(i).team() == OUR_TEAM_ID ) {
      os << "OUR " << 
        (w.situation().playerWithIndex(i).id() - MIN_TEAMMATE_ID + 1);
    }
    else if( w.situation().playerWithIndex(i).team() == THEIR_TEAM_ID ) {
      os << "THEIR " << 
        (w.situation().playerWithIndex(i).id() - MIN_OPPONENT_ID + 1);
    }
    else {
      os << "OTHER ?"; 
    }
    if( w.situation().playerWithIndex(i).isGoalie() ) {
      os << "(G)";
    }
    else {
      os << "(" + Ath98String( globalPlayerTypes->typeOfPlayerWithId( 
          w.situation().playerWithIndex(i).id() ) ) + ")";
    }
    os <<
      ": p(" << w.situation().playerWithIndex(i).pos().x() <<
      ", " << w.situation().playerWithIndex(i).pos().y() <<
      ") s(" << w.situation().playerWithIndex(i).speed().x() <<
      ", " << w.situation().playerWithIndex(i).speed().y() <<
      ") b(" << degFromRadiant( w.situation().playerWithIndex(i).bodyDir() ) <<
      ") h(" << degFromRadiant( w.situation().playerWithIndex(i).faceDir() ) <<
      ")";
  }

  os << "\n";
  return os;
}




