#ifndef _INTENTION_H_
#define _INTENTION_H_

#include <string>
#include <iostream>
#include <list>
#include <algorithm>

#include "../basic/Action.h"

#include "targets/Target.h"
#include "conditions/Condition.h"

#include "IntentionDictionary.h"

/**
   Die Klasse Intention ist die abstrakte Basisklasse (Konstruktoren sind protected)
   von der alle weiteren Intention abgeleitet werden m\"ussen. Jede abgeleitete
   Intention mu\3 eine endeutige Intention-Id besitzten, die sie charakterisiert.
   Ihr mu\3 auch im Intention-Dictionary ein Name, der der Ausgabe dient, zugeordnet
   werden.
   Jede Intention kann eine Liste von Unter-Intentions besitzten, an die Intentionen
   angef\"ugt (\Ref{addIntention}) werden k\"onnen. Diese werden der Reihenfolge nach
   beim execute-Aufruf der Intention abgearbeitet.
   Eine Intention besitzt immer Pre- und Post-Conditions, die in jeweils einer Liste
   organisiert sind. Bei der Ausf\"uhrung einer Intention werden diese Conditions
   gepr\"uft. Zun\"achst werden die Post-Conditions ausgewertet, danach die
   Pre-Conditions.
   Im Konstruktor jeder Intention sollten die Conditions und die Unter-Intentions
   (sofern vorhanden) angelegt und in die jeweilige Liste aufgenommen werden.
   Zum Anlegen der Unter-Intentions sollte ein Methode namens "calculateIntentionOrder"
   benutzt werden.
   Die Methode \Ref{execute} ist die Ausf\"uhrung der Intention. Sie liefert an die
   aufrufende Instanz eine Action zur\"uck, die den Effektoren zum Senden an den
   Soccerserver \"ubergeben wird. Es darf keine undefinierte Action zur\"uckgegeben
   werden. Falls eine Intention oder eine ihrer Unter-Intentions eine leere Action
   (EMPTY_ACTION) liefert, mu\3 diese eine eventuell erforderliche Sonderbehandlung
   selbst durchf\"uhren.
   Besitzt eine Intention eine Liste mit mehreren Unter-Intentions, so wird diese von
   Anfang an durchgegangen und es findet ein Wechsel zur n\"achsten Intention immer
   dann statt, wenn entweder die Post-Conditions der aktuellen erf\"ullt sind, oder
   deren Pre-Condition noch nicht zutriftt. Um durch die Liste zu iterieren, wird der
   Iterator auf das n\"achste Element der Liste weitergesetzt, und die execute-Methode
   rekursiv aufgerufen. Ist die Liste leer (geworden), kann eine Reparatur durchgef\"uhrt
   werden. D.h. es kann eine (oder mehrere bei Fallunterscheidung) Intentionen geben,
   die eine Reparatur versuchen.
   Da die execute-Methode im zeitkritischen Abschnitt des Agenten liegt, sollte im
   Reparaturfall immer zuerst die emergencyController-Routine des Timers aufgerufen
   werden, um festzustellen, ob noch gen\"ugend Zeit vorhanden ist, um die Reparatur
   und ein eventuelles Backtracking durchzuf\"uhren. Ist die Zeit zu knapp, so mu\3
   eine von der globalen EmergencyIntention generierte Default-Action zur\"uckgeliefert
   werden, damit der Agent \"uberhaupt eine Handlung ausf\"uhrt.
   Der Log-Mechanismus der Intentionen macht es notwendig, da\3 jede Intention einen Namen
   im IntentionDictionary besitzt. Die Id der Intention ist dann der Schl\"ussel zum
   zugeh\"origen Namen.
   Die Id einer Intention wird in dieser als static const definiert und ist vom Typ
   IntentonId. Diese Id ist dem Konstruktor der BasisIntention zu \"ubergeben und wird
   in ihr abgelegt. Somit ist es m\"oglich, da\3 die Log-Ausschriften der Intentions
   immr von dem Ausgabeoperator der Basisintention gemacht werden k\"onnen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Intention.h} | [\URL[Implementierung]{src/Intention.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Intention.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/



typedef unsigned short IntentionId;

class Intention;

extern void eraseCondition( Condition *a );
extern void eraseIntention( Intention *i );

class Intention {
 public:
  typedef unsigned short IntentionId;
 protected:
    /** Konstruktoren. */
  Intention( IntentionId aId );
  Intention( IntentionId aId, const Target& aTarget );

 public:
  Intention( const Intention& anIntention );
  

  /** Destruktor, weil virtuelle Funktionen vorhanden sind. */
  virtual ~Intention();

  /** Ausfhren der Intention. Liefert die n"achste auszuf"uhrende Aktion
      zur"uck. */
  virtual Action execute();

  /** Der virtuelle Deliberator -- bestimmt die Intentions */
  virtual void calculateIntentionOrder();
 
  /** Checken der Vorbedingungen. Sollte nur im Ausnahmefall von
      abgeleiteter Klasse redefiniert werden. */
  virtual bool checkPreCond() const;

  /** Checken der Nachbedingungen. Sollte nur im Ausnahmefall von
      abgeleiteter Klasse redefiniert werden. */
  virtual bool checkPostCond() const;

  /** Checken der Break-Bedingungen. Sollte nur im Ausnahmefall von
      abgeleiteter Klasse redefiniert werden. */
  virtual bool checkBreakCond() const;

  /** nur redefinieren, wenn eine Initialisierung bentigt wird 
      Beispiel : FreeKickIntention Rollenspiel wird berechnet ,
      wenn die FreeKickIntention eine andere vorige I. abloest */
  virtual void initialize();

  /** Setzen des Ziel f"ur die Intention. */
  void setTarget( const Target& aTarget );

  /** R"uckgabe des Ziels der Intention. */
  const Target& target() const;

  /** Anf"ugen einer Vorbedingung. */
  void addPreCondition( Condition* aCondition );

  /** Anf"ugen einer Nachbedingung. */
  void addPostCondition( Condition* aCondition );  

  /** Anf"ugen einer Ausnahmebedingung. */
  void addBreakCondition( Condition* aCondition );

  /** Anf"ugen einer Intention. */
  void addIntention( Intention* aIntention );

  /** Die auszuf"uhrende Aktion. */
  Action action();

  /** Liefert einen String mit Klassenbezeichnung fuer das Log zurueck. */
  virtual string idString() const;

  IntentionId id() const;
  IntentionId lastIntentionId() const;

 protected:

  /** Liste der Vorbedingungen. */
  list<Condition*> thePreConditions;
  /** Liste der Nachbedingungen. */
  list<Condition*> thePostConditions;
  /** Liste der Intentions. */
  list<Intention*> theIntentions;
  /** Iterator, der auf die aktuelle Intention zeigt. */
  list<Intention*>::iterator current;

  /** Liste der Ausnahmebedingungen. */
  list<Condition*> theBreakConditions;

  /** Id der zuletzt gerufenen Sub-Intention */
  IntentionId theLastIntentionId;

  int executerCycle;

  /** Setzen der Id. */
  void setId( IntentionId anId );
  /** Setzen der Id, der zuletzt gerufenen Sub-Intention. */
  void setLastIntentionId( IntentionId anId );

  /** @name L"oschen der Listen. */
  //@{
  ///
  void eraseAllPreConditions();
  ///
  void eraseAllPostConditions();
  ///
  void eraseAllIntentions();
  ///
  virtual void eraseAllBreakConditions();
  //@}

  /** Die Id der jeweiligen Intention. */
  IntentionId theId;

  /** Das Ziel der Intention. */
  Target theTarget;

 private:

  /** Die n"achste Aktion, die auszuf"uhren ist. Die Aktion selber wird
      mittels \Ref{execute} berechnet. */
  Action theNextAction;
};

inline IntentionId Intention::id() const{
  return theId;
}

inline void Intention::setLastIntentionId( IntentionId anId ){
  theLastIntentionId = anId;
}

inline IntentionId Intention::lastIntentionId() const{
  return theLastIntentionId;
}

inline string Intention::idString() const{
  return globalIntentionDictionary[ theId ];
}

inline void Intention::initialize(){}

inline Action Intention::action(){
  if( theNextAction.simStep() < globalWorldModel->simStep() ){
    theNextAction = execute();
  }
  return theNextAction;
}

inline void Intention::setTarget( const Target& aTarget ){
  theTarget = aTarget;
}

inline const Target& Intention::target() const{
  return theTarget;
}

inline void Intention::addPreCondition( Condition* aCondition ){
  thePreConditions.push_back( aCondition );
}

inline void Intention::addPostCondition( Condition* aCondition ){ 
  thePostConditions.push_back( aCondition );
}

inline void Intention::addIntention( Intention* aIntention ){
  theIntentions.push_back( aIntention );
}

inline void Intention::addBreakCondition( Condition* aCondition ){
  theBreakConditions.push_back( aCondition );
}

inline void eraseCondition( Condition *a ) {
  delete a;
}

inline void eraseIntention( Intention *i ) {
  delete i;
}

inline void Intention::eraseAllBreakConditions(){
  for_each( theBreakConditions.begin(), theBreakConditions.end(), eraseCondition );
  theBreakConditions.clear();
}

inline void Intention::eraseAllPreConditions(){
  for_each( thePreConditions.begin(), thePreConditions.end(), eraseCondition );
  thePreConditions.clear();
}

inline void Intention::eraseAllPostConditions(){
  for_each( thePostConditions.begin(), thePostConditions.end(), eraseCondition );
  thePostConditions.clear();
} 

inline void Intention::eraseAllIntentions(){
  for_each( theIntentions.begin(), theIntentions.end(), eraseIntention );
  theIntentions.clear();
}

#endif
