#ifndef _BODY_INFOS_H_
#define _BODY_INFOS_H_

#include <stream.h>
#include "../basic/Common.h"
#include "../basic/Ath98String.h"

/**
  Informationen ueber den Zustand des eigenen Koerpers. Dieses Objekt
  enthaelt alle Daten die einer sense_body Nachricht des SoccerServers
  zu entnehmen sind. Diese sind die Sichtqualitaet, der Sichtwinkel,
  der Wert des Stamina, der Wert des Efforts, der Betrag der Geschwindigkeit,
  und die Anzahl der Kicks, Dashs, Turns und Says. Zusaetzlich
  enthaelt dieses Objekt noch die Zeit zu der gerufen wurde.

  {\small {\bf Quelltext}

  [\URL[Header]{src/BodyInfos.h} | [\URL[Implementierung]{src/BodyInfos.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: BodyInfos.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $
  *
  **************************************************
  \end{verbatim} }*/
class BodyInfos
{
public:
  /**
    Default-Konstruktor. Initialisierung mit einem Ath98String. */
  BodyInfos( const Ath98String& information );

  /** Konstrukter fuer Template-Klassen */
  BodyInfos();

  /** Die Zeit zu der die Informationen ueber den eigenen Koerper gesendet
    wurden. */
  long time() const;

  int viewQuality() const;

  int viewWidth() const;

  double stamina() const;

  double effort() const;

  double amountOfSpeed() const;

  double directionOfSpeed() const;

  double headAngle() const;

  int kicks() const;

  int dashs() const;

  int turns() const;

  int says() const;

  int turnNecks() const;

  int moves() const;

  int catches() const;

  // HACK fuer GoalKickAfterCatchPositionAnalyzer!!!
  // ging nicht anders, wirklich. =(
  static int staticMoves;
  static int staticCatches;	

  int changeViews() const;
  /** Returns the corresponding count for the given actionId if valid.
      Otherwise UNDEF_INT is returned.
  */
  const int countForActionId( const int anActionId ) const;

  long int count() const;

  void setCount( long int aCount );

  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const BodyInfos& );

  /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal. */
  Ath98String toString() const;

  /** kompakte Ausgabe auf einen Stream fr constant Log */
  void toStampStream( ostream& ) const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
  BodyInfos* copy() const;

private:
  /** Setzt alle Daten, die man dem Ath98String entnehmen kann. */
  void parseBodyInfos(const Ath98String& information);

  /** Zeit zu der die Informationen ueber den eigenen Koerper gesendet wurde. */
  long theTime;

  int theViewQuality;
  int theViewWidth;
  double theStamina;
  double theEffort;
  double theAmountOfSpeed;
  double theDirectionOfSpeed;
  double theHeadAngle;
  int theKicks;
  int theDashs;
  int theTurns;
  int theSays;
  int theTurnNecks;
  int theMoves;
  int theCatches;
  int theChangeViews;
  /** Nummer die aussagt, die wievielte BodyInfo dieses Object ist. */
  long int theCount;
};

inline long BodyInfos::time() const {
  return theTime;
}

inline int BodyInfos::viewQuality() const {
  return theViewQuality;
}

inline int BodyInfos::viewWidth() const {
  return theViewWidth;
}
  
inline double BodyInfos::stamina() const {
  return theStamina;
}

inline double BodyInfos::effort() const {
  return theEffort;
}
  
inline double BodyInfos::amountOfSpeed() const {
  return theAmountOfSpeed;
}

inline double BodyInfos::directionOfSpeed() const {
  return theDirectionOfSpeed;
}

inline double BodyInfos::headAngle() const {
  return theHeadAngle;
}

inline int BodyInfos::kicks() const {
  return theKicks;
}
  
inline int BodyInfos::dashs() const {
  return theDashs;
}

inline int BodyInfos::turns() const {
  return theTurns;
}
  
inline int BodyInfos::says() const {
  return theSays;
}

inline int BodyInfos::turnNecks() const {
  return theTurnNecks;
}

inline int BodyInfos::moves() const {
  return theMoves;
}

inline int BodyInfos::catches() const {
  return theCatches;
}

inline int BodyInfos::changeViews() const {
  return theChangeViews;
}

inline const int BodyInfos::countForActionId( const int anActionId ) const {
  switch( anActionId ) {
  case DASH_ACTION:
    return( dashs() );
  case TURN_ACTION:
     return( turns() );
  case KICK_ACTION:
    return( kicks() );
  case TURN_NECK_ACTION:
    return( turnNecks() );
  case SAY_ACTION:
    return( says() );
  case MOVE_ACTION:
    return( moves() );
  case CATCH_ACTION:
    return( catches() );
  case CHANGE_VIEW_ACTION:
    return( changeViews() );
  default:
    return( UNDEF_INT );
  }
}

inline long int BodyInfos::count() const {
  return theCount;
}

inline void BodyInfos::setCount( long int aCount ) {
  theCount = aCount;
}

inline ostream& operator<< (ostream& os, const BodyInfos& b ) {
  os <<
    "( BodyInfos time: " <<  b.time() << 
    ", viewQuality: " <<  b.viewQuality() << 
    ", viewWidth: " <<  b.viewWidth() <<
    ", stamina: " <<  b.stamina() <<
    ", effort: " <<  b.effort() <<
    ", amountOfSpeed: " <<  b.amountOfSpeed() <<
    ", headAngle: " << b. headAngle() <<
    ", kicks: " <<   b.kicks() <<
    ", dashs: " <<  b.dashs() <<
    ", turns: " <<  b.turns() <<
    ", says: " <<  b.says() <<
    ", turnNecks: " <<  b.turnNecks() <<  
    " )" ;   
  return os;
}

inline void BodyInfos::toStampStream( ostream& os) const {

  os << " " << BODY_INFOS_MARK 
     << " " <<  viewQuality()  
     << " " <<  viewWidth() 
     << " " <<  stamina() 
     << " " <<  effort() 
     << " " <<  amountOfSpeed() 
     << " " <<  directionOfSpeed()
     << " " <<  headAngle() 
     << " " <<  kicks() 
     << " " <<  dashs() 
     << " " <<  turns() 
     << " " <<  says() 
     << " " <<  turnNecks() ;        
}

#endif







