#ifndef _COMMUNICATOR_H_
#define _COMMUNICATOR_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string>
#include <unistd.h>
#include <string>
#include <vector>

#include "../basic/Common.h"
#include "../basic/Ath98String.h"
#include "../basic/Exception.h"
#include "../basic/Configuration.h"

/**
  Der UDP/IP-level Kommunikationsverwalter. 
  Der Communicator sorgt f"ur die Kommunikation mit dem SoccerServer auf der
  UDP/IP Ebene. Er schickt Ath98Strings an den Server und liest blockierend 
  Ath98Strings vom Server.

  Von der Klasse wird nur ein Objekt erzeugt. Die Klasse ist nicht zur
  Vererbung geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Communicator.h} | [\URL[Implementierung]{src/Communicator.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Communicator.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/

class Communicator
{
public:
  /** Interne Fehlerklasse. */
  class CommunicatorException: public Exception {
  public:
    CommunicatorException(const Ath98String& aDescription )
      : Exception( aDescription ) {}
  };

  /** Exception die ansagt, da"s keine Nachricht erhalten wurde. 
   */
  class NothingReceived: public Exception {
  public:
    NothingReceived() : Exception() {}
  };  
  
  /** Verbindungsaufbau.
    Aufbau einer UDP/IP Verbindung.
    @param hostName Der Name des Rechners, auf dem der SoccerServer laeuft
    @param port Der Port, an dem der SoccerServer auf Spieler wartet. 
    @exception CommunicatorException Noch nicht genauer spezifiziert. */
  Communicator(const Ath98String& hostName, int port );
  
  /** Verbindungsabbau.
    Die Verbindung zum SoccerServer wird beendet. */
  ~Communicator();
  
  /** Schicken einer Nachricht an den Server.
    Die Nachricht wird sofort per UDP an den Server geschickt.
    @param theMessage Ath98String, der die Nachricht enthaelt
    @exception CommunicatorException Ath98String konnte nicht abgesendet werden */
  void sendMessage(const Ath98String& theMessage);
  
  /** Liefert die naechste Nachricht, die vom Server kommt.
    Ggf. wird gewartet, bis eine Nachricht eintrifft.
    @param seconds Maximale Wartezeit in Sekunden. Default ist WAIT_FOREVER	
    @exception NothingReceived Keine Nachricht eingetroffen. */
  Ath98String getNextMessage(int seconds = WAIT_FOREVER);

  /** Liefert die zuletzt erhaltene Nachricht, die vom Server kamm.
    Ggf. wird gewartet, bis eine Nachricht eintrifft.
    @exception CommunicatorException Keine Nachricht eingetroffen. */
  Ath98String getLastMessage();  
  
  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;
  
  /** Liefert die gesendeten Actionen und l\"oscht. */
  void sendedActionsToStampStream( ostream& os );

private:
  int ownSocket;
  struct sockaddr_in hostAddress;
  unsigned int lengthOfHostAddress;
  bool communicationAvailable;
  
  void openConnection(const Ath98String& hostName, int port);
  void closeConnection();

  /** Array mit den gesendeten Aktionen. */
  vector< string > theSendedActions;

  /** String der die Kurzkennung der Klasse enthlt */
  string theIdString;
};


inline string Communicator::idString() const {
  return theIdString;
}

inline void Communicator::sendedActionsToStampStream( ostream& os ) {
    os << SENDED_ACTIONS_MARK;
  for( vector< string >::const_iterator i = theSendedActions.begin();
       i != theSendedActions.end() ; ++i ) {
    os << " " << *i;
  }
  theSendedActions.clear();
}

#endif
