#include "Effectors.h"

#include "../basic/Log.h"
#include "../basic/Timer.h"

#include "../worldmodel/ModelledBall.h"
#include "../worldmodel/WorldModelInterface.h"
#include "../worldmodel/Neck.h"

/**
 * Aktueller Stand
 *
 *   $RCSfile: Effectors.cc,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 2001/07/31 22:23:09 $
 *   $Author: rami $
 *
 */

Effectors::Effectors( Communicator* aCommunicator )
  : theCommunicator( aCommunicator ) {
    hasEnteredField = NO;
    theQueue = vector<Action>(20);
    indexOfCurrentAction = 0;
    theLastMainAction = Action();
    theLastSubAction = Action();
    theIdString = "Effectors";
 }

void Effectors::enterField( const Ath98String& aTeamName, bool isGoalie ) {
  Ath98String message;
  Ath98String answer;

  if ( hasEnteredField ) {
    WARNING( Log::EFFECTORS, "Field already entered. Ignoring command." );
    return;
  }

  if(isGoalie) {
    message = "(init " + aTeamName + " (version 7.00) (goalie))";
  }
  else {
    message = "(init " + aTeamName + " (version 7.00))";
  }

  theCommunicator->sendMessage( message );
  try {
    answer = theCommunicator->getNextMessage(
        globalConf->communicationTimeout );
  }
  catch( Communicator::NothingReceived ) {
    throwException( EffectorsException( "Server didn't respond" ) );
  }
  if( answer.beginsWith( "(error ") ) {
    throwException(
        EffectorsException( answer ) );
  }

  char internalSide;
  int playerNumber;
  sscanf (answer.cString(), "(init %c %i ", &internalSide, &playerNumber);
  globalWorldModel->setTeamName( aTeamName );
  globalWorldModel->setSide( (internalSide == 'l') ?
			     LEFT_SIDE_ID : RIGHT_SIDE_ID );

  globalWorldModel->setSimStep( 0 );
  globalWorldModel->setGameState( WorldModelInterface::BEFORE_KICK_OFF );
  globalWorldModel->setInitiative( LEFT_SIDE_ID );
  globalWorldModel->setPlayerNumber( playerNumber, isGoalie );

  if ( isGoalie == YES ) {
    if ( globalWorldModel->role() != GOALIE_ROLE ) {
      throwException( EffectorsException(
          "Role of Goalie is not GOALIE_ROLE" ) );
    }
  }
  else {
    if ( globalWorldModel->role() == GOALIE_ROLE ) {
      globalWorldModel->setRole( ASSISTANT_GOALIE_ROLE );
      throwException( EffectorsException(
          "Goalie didn't connect as goalie." ) );
      //WARNING( Log::EFFECTORS, Ath98String( "Goalie didn't connect as goalie." ) +
      //    "Setting role to ASSISTANT_GOALIE_ROLE.\n" );
    }
  }
  hasEnteredField = YES;
}

void Effectors::reEnterField( const Ath98String& aTeamName, int aPlayerNumber, bool isGoalie ) {
  Ath98String message;  
  Ath98String answer;  
   
  if ( hasEnteredField ) {
    WARNING( Log::EFFECTORS, "Field already entered. Ignoring command." );  
    return;  
  } 
   
  message = "(reconnect " + aTeamName + " " + aPlayerNumber + ")";  
   
  theCommunicator->sendMessage( message );  
  try { 
    answer = theCommunicator->getNextMessage( 5 );  
  } 
  catch( Communicator::NothingReceived ) { 
    throwException( EffectorsException( "Server didn't respond" ) );
  }
  if( answer.beginsWith( "(error" ) ) {
    throwException( EffectorsException( "Server refused reconnect." ) );
  }

  char internalSide;
  sscanf (answer.cString(), "(reconnect %c ", &internalSide);

  globalWorldModel->setPlayerNumber( aPlayerNumber, isGoalie );
  globalWorldModel->setTeamName( aTeamName );
  globalWorldModel->setSide( (internalSide == 'l') ?
			     LEFT_SIDE_ID : RIGHT_SIDE_ID );
  globalWorldModel->setSimStep( globalConf->halfTime );
  globalWorldModel->setGameState( WorldModelInterface::BEFORE_KICK_OFF );
  globalWorldModel->setInitiative( RIGHT_SIDE_ID );
  hasEnteredField = YES;
}

void Effectors::leaveField() { 
  if ( !hasEnteredField ) { 
    WARNING( Log::EFFECTORS, "Not on field. Ignoring command." );  
    return;  
  } 
   
  theCommunicator->sendMessage( "(bye)" );  
  hasEnteredField = NO;  
} 

bool Effectors::checkAndResendAction( const BodyInfos& aBodyInfo ) {
  bool wasActionLost = NO;
  Action lostAction;
  INFO( Log::EFFECTORS  ,  4  ,  idString() , globalLog->infoStream << 
	" ENTER: Effectors::checkAndResendAction\n"  ) ;

  switch( theLastMainAction.id() ) {
  case DASH_ACTION:
    if( aBodyInfo.dashs() == theLastMainAction.actionCount() ) {
      wasActionLost = YES;
    }
    break;
  case TURN_ACTION:
    if( aBodyInfo.turns() == theLastMainAction.actionCount() ) {
      wasActionLost = YES;
    }
    break;
  case KICK_ACTION:
    if( aBodyInfo.kicks() == theLastMainAction.actionCount() ) {
      wasActionLost = YES;
    }
    break;
  }
  if( wasActionLost ) {
    lostAction = theLastMainAction;
  }
  else if( theLastSubAction.id() == TURN_NECK_ACTION && 
           aBodyInfo.turnNecks() == theLastSubAction.actionCount() ) {
    wasActionLost = YES;
    lostAction = theLastSubAction;
  }

  if( wasActionLost ) {
    WARNING( Log::EFFECTORS, "Action: " << lostAction.toString() << 
             "was lost. Resending...\n" );
    performAction( lostAction );
  }
  INFO( Log::EFFECTORS  ,  4  ,  idString() , globalLog->infoStream << 
	" LEAVE: Effectors::checkAndResendAction\n"  ) ;
  return wasActionLost;
}

void Effectors::performAction( Action anAction, const BodyInfos& aBodyInfo )
{
  switch( anAction.id() ) {
  case DASH_ACTION:
    anAction.setActionCount( aBodyInfo.dashs() );
    break;
  case TURN_ACTION:
    anAction.setActionCount( aBodyInfo.turns() );
    break;
  case KICK_ACTION:
    anAction.setActionCount( aBodyInfo.kicks() );  
    break;
  case TURN_NECK_ACTION:
    anAction.setActionCount( aBodyInfo.turnNecks() );
    break;
  case SAY_ACTION:
    anAction.setActionCount( aBodyInfo.says() );
    break;
  }

  performAction( anAction);
}


void Effectors::performAction( Action& anAction ) {

// cout << anAction << endl;
  INFO( Log::EFFECTORS, 2, idString(), globalLog->infoStream <<
	"    ENTER Effectors::performAction\n" );

  if ( anAction.isValid() == NO ) {
    throwException( EffectorsException( "You tried to let me do the invalid action " + anAction.toString() ) );  
  } 
  
  // Zuletzt abgesendetes Kommando fuer eine eventuelle Nachsendung speichern.
  switch( anAction.id() ) {
    case DASH_ACTION:
    case TURN_ACTION:
    case KICK_ACTION:
      theLastMainAction = anAction;
      break;
    case TURN_NECK_ACTION:
      theLastSubAction = anAction;
      break;
  }

  if ( anAction.simStep() != globalTimer->soccerServerSimStep() && anAction.isMainAction() ) {
    WARNING( Log::EFFECTORS, "Sending an action that was calculated for "
              + Ath98String( anAction.simStep() ) + " at simStep "
              + Ath98String(  globalTimer->soccerServerSimStep() ) + "\n" );
    Action modifiedAction( anAction );
    modifiedAction.setSimStep( globalTimer->soccerServerSimStep() );
    globalWorldModel->registerExecution( modifiedAction );
  }
  else {
    globalWorldModel->registerExecution( anAction );
  }

  if( anAction.id() != EMPTY_ACTION ) {
    theCommunicator->sendMessage( anAction.asCommandAth98String() );
//if( anAction.id() == TURN_ACTION && anAction.actionCount() == 5 ) exit(0);
//cerr << " Action at " << anAction.simStep() << "   " << anAction.asCommandAth98String() << endl;
//if (globalWorldModel->simStep() == 5) exit(0);
      // der nacken erfhrt das gesendete Kommando
    globalNeck->registerSendedAction( anAction );
  }
//else cerr << "Effectors::performAction EMPTY_ACTION aufgerufen\n";

  INFO( Log::EFFECTORS, 2, idString(), globalLog->infoStream <<
	"    LEAVE Effectors::performAction\n" );
} 

Ath98String Effectors::toString() const 
{ 
  int numberOfActions = theQueue.size();
 
  Ath98String output = "( Effectors actions: ";  
  for( int i = indexOfCurrentAction;  i < numberOfActions;  i++ ) { 
    output = output + "\n" + Ath98String( i ) + ".: " +  
      theQueue[i].toString();
  } 
  output = output + " )\n";  
  return output;  
} 
 
