#ifndef _EFFECTORS_H_
#define _EFFECTORS_H_

#include <stdio.h>
#include <stream.h>
#include <string>

#include "Communicator.h"
#include "../basic/Ath98String.h"
#include "../basic/Action.h"
#include "../basic/Exception.h"
#include "BodyInfos.h"

/**
  Handlungsausfuehrende Komponente. 
  Die Basisaktionen, die der Server zur Verfuegung stellt, sind hier gekapselt.
  Effectors uebernimmt ausserdem das Queueing der Kommandos und das
  Benachrichtigen des Weltmodells ueber tatsaechlich abgeschickte Aktionen.

  Alle Aktionen, zu denen es Action-Objekte gibt, koennen mit queueAction 
  ausgefuehrt werden. Ausserdem koennen Plaene, also Abfolgen von Aktionen mit
  queuePlan in die Warteschlange gehaengt werden.

  Die Aktionen enterField, reEnterField und leaveField muessen direkt
  ueber den Aufruf der gleichnamigen Methode ausgeloest werden.

  Von der Klasse wird nur ein Objekt erzeugt. 
  Die Klasse ist nicht zur Vererbung geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Effectors.h} | [\URL[Implementierung]{src/Effectors.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Effectors.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class Effectors
{
public:
  /** Interne Fehlerklasse. */
  class EffectorsException: public Exception 
  {
  public:
    EffectorsException(const Ath98String& aDescription )
      : Exception( aDescription ) {}
  };
  
  /** Interne Fehlerklasse. Wird aufgerufen, wenn beim Wunsch Kommandos
    zu senden keine Kommandos mehr in der Queue sind. */
  class NothingToSendException: public Exception
  {
  public:
    NothingToSendException() :  Exception() {}
  };
    
  /** Erzeugt ein neues Effectors-Objekt. 
    @param aCommunicator Der Communicator, ueber den Effectors die
    Aktionen verschickt. */
  Effectors( Communicator* aCommunicator );
  
  /** Betritt das Spielfeld zur ersten Halbzeit auf der Seite des Teams
    aTeamName. Ist das nicht moeglich, wird eine EffectorsException erzeugt. Im
    Weltmodell werden eingetragen:
    \begin{itemize}
    \item Die Spielernummer, die der Server beim Anmelden zurueckschickt
    \item Die Seite, auf der der Spieler spielt
    \item Der Teamname
    \end{itemize}
    Das Weltmodell muss also vor Aufruf dieser Methode bereits initialisiert
    sein. */
  void enterField( const Ath98String& aTeamName, bool goalie = NO );

  /** Betritt das Spielfeld zur zweiten Halbzeit auf der Seite des Teams
    aTeamName. Ist das nicht moeglich, wird eine EffectorsException erzeugt. Im
    Weltmodell werden eingetragen:
    
    \begin{itemize}
    \item Die bereits bekannte Spielernummer,
    \item Die Seite, auf der der Spieler spielt
    \item Der Teamname.
    \end{itemize}
    Das Weltmodell muss also vor Aufruf dieser Methode bereits initialisiert
    sein. */
  void reEnterField( const Ath98String& aTeamName, int aPlayerNumber, bool isGoalie = NO );
	
  /** Veranlasst das Verlassen des Feldes. */
  void leaveField();
  
  /** Fuehrt eine Aktion sofort aus, ohne sie in die Warteschlange zu haengen.
    Die Erstellung des korrekten Ath98Strings fuer
    den Server liegt in der Verantwortung der Aktion. Es findet keinerlei
    Fehlerueberpruefung des Ath98Strings statt. Die Aktion wird allerdings vorher
    gefragt, ob sie meint, einen gueltigen CommandString erzeugen zu koennen.
    Wenn ja, wird er gesendet, wenn nicht, wird eine EffectorsException
    ausgeloest.  */
  void performAction( Action& anAction );
  
  /** Ruft die Methode \Ref{performAction} auf nachdem es die
      actioncounts fuer die anAction gesetzt hat. */
  void performAction( Action anAction, const BodyInfos& aBodyInfo );

  /** Test und falls noetig erneutes Abschicken der letzten Kommandos.
      In Frage kommt eines der Kommandos dash, turn, kick und das Kommando
      turn_neck. Neu abgesendet wird ein Kommando, wenn der Zaehler dieses
      Kommandos in den uebergebenen bodyInfos dem Zaehler eines der
      zuletzt gesendeten Kommandos entspricht. 
      Falls beide Kommandos (eine MainAction und eine SubAction) nicht beim
      SoccerServer angekommen sind, wird nur die MainAction erneut gesendet.
      @return YES wenn erneut gesendet wurde, sonst NO
*/
  bool checkAndResendAction( const BodyInfos& aBodyInfo );

  /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal. */
  Ath98String toString() const;  
	
  /** Ausgabe auf einen String-Stream */
  friend ostream& operator<< (ostream&, const Effectors& );

  /** kompakte Ausgabe auf einen Stream fr constant Log */
  void toStampStream( ostream& ) const;

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;

private:
  /** Pointer auf den Communicator des Agenten. */
  Communicator *theCommunicator;
  
  /** Ist das Feld betreten? Die Menge der erlaubten Aktionen ist davon
    abhaengig. */
  bool hasEnteredField;

  /** Zuletzt gesendete Hauptaktion. Speicherung um bei Nichtempfang vom
      SoccerServer diese Kommando nachsimmulieren zu koennen. */
  Action theLastMainAction;

  /** Zuletzt gesendete Nebenaktion. Speicherung um bei Nichtempfang vom
      SoccerServer diese Kommando nachsimmulieren zu koennen. */
  Action theLastSubAction;
  
  /** Die Aktionsqueue. */
  vector<Action> theQueue;
  
  /** Der Index der gegenwaertig aktuellen Aktion. */
  unsigned int indexOfCurrentAction;

  /** String der die Kurzkennung der Klasse enthlt */
  string theIdString;

};

inline ostream& operator<< (ostream& os, const Effectors& Effectors )
{
  unsigned int numberOfActions = Effectors.theQueue.size();
  os << 
    "( Effectors actions: ";
  for( unsigned int i = Effectors.indexOfCurrentAction; i < numberOfActions; i++ ) {
    os << "\n" << i  << ".: " <<
      Effectors.theQueue[i];
  }
  os <<  " )\n";
  return os;
}

inline void Effectors::toStampStream( ostream& os) const {
  int numberOfActions = theQueue.size();
  os << " " << QUEUESTART_MARK;
  for( int i = indexOfCurrentAction; i < numberOfActions; i++ ) {
    theQueue[i].toStampStream( os );
  }
  os <<  "\n";
}

inline string Effectors::idString() const
{
  return theIdString;
}

#endif

