#ifndef _CONFIGUREDPLAYER_H_
#define _CONFIGUREDPLAYER_H_

#include <stream.h>

#include "ModelledObject.h"
#include "MovingPlayer.h"

#include "../basic/Configuration.h"
#include "../basic/Common.h"
#include "../basic/Exception.h"
#include "../basic/Vector.h"

#include "PlayerTypes.h"

/**
  Die interne Repraesentation eines Spielers im Weltmodell.
  Ein ConfiguredPlayer hat zusaetzlich zu den Eigenschaften eines ModelledObject
  noch eine Blickrichtung und eine Kraftreserve. Die Kraftreserve setzt sich
  aus
  den Werten stamina, effort und recovery zusammen. Dazu kommen entsprechende
  Verl\"asslichkeitsinformationen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/ConfiguredPlayer.h} | [\URL[Implementierung]{src/ConfiguredPlayer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ConfiguredPlayer.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class ConfiguredPlayer : public ModelledObject
{

public:

  /** @name Interne Fehlerklassen. */
  //@{
    /** Allgemeine Fehlerklasse. */
    class ConfiguredPlayerException: public ModelledObjectException {
    public:
      ConfiguredPlayerException(const Ath98String& aDescription )
        : ModelledObjectException( aDescription ) {}
    };
  //@}
 
    ConfiguredPlayer( int id  = UNDEF_ID,
		      const Vector& aPos = Vector(), 
		      const Vector& aSpeed = Vector(),
		      double stamina = globalConf->playerStaminaMax,
		      double effort = 1.0,
		      double recovery = 1.0,
		      const Confidence& aPosConfidence = Confidence(),
          const Confidence& aSpeedConfidence = Confidence() );

    ConfiguredPlayer( const MovingPlayer& aMovingPlayer,
		      double stamina = globalConf->playerStaminaMax,
		      double effort = 1.0,
		      double recovery = 1.0,
          const Confidence& aPosConfidence = Confidence(),
          const Confidence& aSpeedConfidence = Confidence() );

 /** Liefert den Stamina-Wert. */
    double stamina() const;
    
    /** Liefert den Effort-Wert. */
    double effort() const;
    
    /** Liefert den Recovery-Wert. */
    double recovery() const;
    

  /** Setzt Stamina-Modell. 
      @param newStamina Die Kraftreserve. Gueltige Werte: 0 bis theStaminaMax
      @param newEffortFactor Der Leistungsfaktor.
      Gueltige Werte: theEffortMin bis 1.0
      @param newRecoveryFactor Der Erholungsfaktor.
      Gueltige Werte: theRecoveryMin bis 1.0
      */
    void setStaminaValues( double newStamina = UNDEF_DOUBLE,
      double newEffortFactor = UNDEF_DOUBLE,
      double newRecoveryFactor = UNDEF_DOUBLE );

    /** Setzt den Stamina-Wert. */
    void setStamina( double newStamina );

    /** Setzt den Effort-Wert. */
    void setEffort( double newEffortFactor );

    /** Setzt die Recovery. */
    void setRecovery( double newRecoveryFactor );

  /** @name Dash-Berechnung */
  //@{
    /** Es wird die tatsaechlich verfuegbare Kraft fuer einen dash
      (effectiveDashPower) ausgerechnet. Diese ist abhaengig vom Effort und vom
      noch zur Verfuegung stehenden Stamina.
      @param power Die Kraft, mit der man beschleunigen will.
      @return effectiveDashPower Die Kraft, mit der man tatsaechlich 
      beschleunigt.  */
    double effectiveDashPower( double power ) const;

    /** Es wird die Kraft bestimmt, die dem SoccerServer geschickt werden muss,
      damit man effektiv mit der Kraft power beschleunigt. 
      Diese Funktion ist also genau die Umkehrfunktion von effectiveDashPower.
      Es gilt also: dashPowerForSend(effectiveDashPower( power )) == power,
      wenn power normalisiert ist (im definierten Bereich liegt). Liegt das
      zu sendene Kommando nicht im definierten Bereich, wird normalisiert und
      eine Warnung ins Log geschrieben.
      @param power Die Kraft mit der der Spieler beschleunigen soll.
      @return powerForSend Die Kraft die man den SoccerServer senden muss. */
    double dashPowerForSend( double power ) const;

    
    /** Normalisiert die Kraft fuer einen dash. 
      @exception ModelledPlayerException falls der Power-Wert undefiniert ist */
    double normalizeDashPower(double power) const;
  //@}

   
  /** @name Turn-Berechnung */
  //@{
    /** Es wird die tatsaechlich ausgefuehrte Drehung
      (effectiveTurnAngle) ausgerechnet. Diese ist abhaengig von der aktuellen
      Geschwindigkeit des Spielers.
      @param angle Der Winkel, um die man sich drehen will.
      @return effectiveTurnAngle Der Winkel, um den man sich tatsaechlich
      dreht.  */
    double effectiveTurnAngle( double angle ) const;
    
    /** Es wird der Winkel bestimmt, der dem SoccerServer geschickt werden muss,
      damit sich der Spieler effektiv um den angegebenen Winkel dreht.
      Diese Funktion ist als genau die Umkehrfunktion von effectiveTurnAngle.
      Es gilt also: turnAngleForSend(effectiveTurnAngle( angle )) == angle,
      wenn angle normalisiert ist (im definierten Bereich liegt). Liegt das
      zu sendene Kommando nicht im definierten Bereich, wird normalisiert
      und eine Warnung ins Log geschrieben.
      @param angle Der Winkel um den sich der Spieler drehen soll.
      @return angleForSend Den Winkel, den man dem Server uebermitteln muss. */
    double turnAngleForSend( double angle ) const;

    /** Formel mit der der Drehwinkel modifiziert wird */
    double turnAngleMult() const;

    /** Baut die Kraftreserven wieder auf. 
	@param simSteps Die Zahl der Simulationsschritte, die zur Erholung zur
	Verfuegung stehen. Wird von simulateSteps() aufgerufen. */
    void recover( int simSteps = 1 );
    
    
    /** Liefert den vom Betrag her maximalen sinnvollen Turn um dessen Winkel
      sich der Spieler in der gegebenen Situation drehen kann. */
    double maxEffectiveTurnAngle() const;

    /** Normalisiert die Drehung fuer einen turn. 
        angle ist ein Winkel im 2001-System! */
    double normalizeTurnAngle(double angle) const;
    
  //@}
    
 /** @name Konfigurierbare Werte erfragen */
  //@{
    ///
    double size() const;
    ///
    double maxSpeed() const;
    ///
    double decay() const;
    ///
    double staminaIncMax() const;
    ///
    double staminaMax() const;
    ///
    double recoverDecThr() const;
    ///
    double recoverDec() const;
    ///
    double recoverMin() const;
    ///
    double effortDecThr() const;
    ///
    double effortDec() const;
    ///
    double effortIncThr() const;
    ///
    double effortInc() const;
    ///
    double effortMin() const;
    ///
    double effortMax() const;
    ///
    double dashPowerRate() const;
    ///
    double maxPower() const;
    ///
    double minPower() const;
    ///
    double maxMoment() const;
    ///
    double minMoment() const;
    ///
    double maxNeckMoment() const;
    ///
    double minNeckMoment() const;
    ///
    double maxNeckAngle() const;
    ///
    double minNeckAngle() const;
    ///
    double normVisibleAngle() const;
    ///
    double inertiaMomentOfTurn() const;
    ///
    double kickableMargin() const;
    ///
    double kickRange() const;
  //@}

    
  /** @name Konfigurierbare Werte setzen */
  //@{
    /** liest die Werte aller Klassenvariablen aus dem Konfigurationsfile 
      ein. */
    void setClassVars();

    ///
    void setStaminaMax(double aStaminaMax);
    
    ///
    void setRecoverDecThr(double aRecoverDecThr);
    
    ///
    void setRecoverDec(double aRecoverDec);
    
    ///
    void setRecoverMin(double aRecoverMin);
    
    ///
    void setEffortDecThr(double aEffortDecThr);
    
    ///
    void setEffortDec(double aEffortDec);
    
    ///
    void setEffortIncThr(double aEffortIncThr);
    
    ///
    void setEffortInc(double aEffortInc);
    
    ///
    void setMaxPower(double aMaxPower);
    
    ///
    void setMinPower(double aMinPower);
    
    ///
    void setMaxMoment(double aMaxMoment);
    
    ///
    void setMinMoment(double aMinMoment);
    
    ///
    void setMaxNeckMoment(double aMaxNeckMoment);
    
    ///
    void setMinNeckMoment(double aMinNeckMoment);

    ///
    void setMaxNeckAngle(double aMaxNeckAngle);
    
    ///
    void setMinNeckAngle(double aMinNeckAngle);

    ///
    void setNormVisibleAngle(double aNormVisibleAngle);

    ///
    void setPlayerType(int aPlayerType);
    
  //@}
  
private:
  ///
  static double theStaminaMax;
  ///
  static double theRecoverDecThr;
  ///
  static double theRecoverDec;
  ///
  static double theRecoverMin;
  ///
  static double theEffortDecThr;
  ///
  static double theEffortDec;
  ///
  static double theEffortIncThr;
  ///
  static double theEffortInc;
  ///
  static double theMaxPower;
  ///
  static double theMinPower;
  ///
  static double theMaxMoment;
  ///
  static double theMinMoment;
  ///
  static double theMaxNeckMoment;
  ///
  static double theMinNeckMoment;
  ///  
  static double theMaxNeckAngle;
  ///
  static double theMinNeckAngle;
  ///
  static double theNormVisibleAngle;
 
 protected:
  ///
  static short isSetOnce;

  /** Der bestimmte oder vermutete PlayerType des Spielers */
  int thePlayerType;
  
  /** Die aktuelle Kraftreserve. Sie sinkt beim Beschleunigen und beim Aufruf
    von usePower. Die Kraftreserve liegt anfangs bei MAX_STAMINA. */
  double theStamina;

  /** Leistungsfaktor, der aktuelle Leistungsfaehigkeit repraesentiert. Er gibt
    an, wieviel Kraft in einem Simulationsschritt tatsaechlich zur Verfuegung
    steht. Er sinkt, wenn die Kraftreserve unter einen Schwellwert faellt und
    steigt wieder, wenn sich der Spieler praktisch ganz erholt hat. */
  double theEffort;
  
  /** Erholungsfaktor, der die Langzeitausdauer repraesentiert. Er gibt an, wie
    stark die Kraftreserve in einer Ruhepause wieder ansteigt. Er wird waehrend
    eines Spiels nicht erhoeht und sinkt, wenn die Kraftreserve unter einen
    Schwellwert faellt. */
  double theRecovery;  

 // Methoden

  /** Normalisiert den Stamina-Wert. */
  double normalizeStamina(double stamina) const;

  /** Normalisiert den Effort-Wert. */
  double normalizeEffort(double effort) const;

  /** Normalisiert den Recovery-Wert. */
  double normalizeRecovery(double recovery) const;

};

inline double ConfiguredPlayer::normalizeStamina(double stamina) const 
{
  if( stamina == UNDEF_DOUBLE ) {
    return staminaMax();
  }
  else if( stamina < 0 ) {
    return 0;
  }
  else if( stamina > staminaMax() ) {
    return staminaMax();
  }
  else{
    return stamina;
  }
}

inline double ConfiguredPlayer::normalizeEffort(double effort) const 
{
  if( effort == UNDEF_DOUBLE ) {
    return effortMax();
  }
  else if( effort < effortMin() ) {
    return effortMin();
  }
  else if( effort > effortMax() ) {
    return effortMax();
  }
  else{
    return effort;
  }
}

inline double ConfiguredPlayer::normalizeRecovery(double recovery) const 
{
  if( recovery == UNDEF_DOUBLE ) {
    return 1.0;
  }
  else if( recovery < recoverMin() ) {
    return recoverMin();
  }
  else if( recovery > 1.0 ) {
    return 1.0;
  }
  else {
    return recovery;
  }
}

inline double ConfiguredPlayer::normalizeDashPower(double power) const 
{
  if( power == UNDEF_DOUBLE ) {
    throwException( ConfiguredPlayerException( "undef power value" ) );
  }
  else if( power < minPower() ) {
    return minPower();
  }
  else if( power > maxPower() ){
    return maxPower();
  }
  else {
    return power;
  }
}

inline double ConfiguredPlayer::normalizeTurnAngle(double angle) const 
{
  angle = normalizeAngle( angle );
  
  if( angle > Ath98_PI ) angle -= 2.0*Ath98_PI;

  if( angle < minMoment() ) {
    angle = minMoment();
  }
  else if( angle > maxMoment() ) {
    angle = maxMoment();
  }
  
  return (angle < 0.0) ? normalizeAngle( angle + 2.0 * Ath98_PI) : angle;
}


inline double ConfiguredPlayer::size() const 
{
  return globalPlayerTypes->playerSize(thePlayerType);
}

inline double ConfiguredPlayer::maxSpeed() const 
{
  return globalPlayerTypes->playerMaxSpeed(thePlayerType);
}

inline double ConfiguredPlayer::decay() const 
{
  return globalPlayerTypes->playerDecay(thePlayerType);
}

inline double ConfiguredPlayer::staminaIncMax() const 
{
  return globalPlayerTypes->staminaIncMax(thePlayerType);
}

inline double ConfiguredPlayer::staminaMax() const 
{
  return ConfiguredPlayer::theStaminaMax;
}

inline double ConfiguredPlayer::recoverDecThr() const 
{
  return ConfiguredPlayer::theRecoverDecThr;
}

inline double ConfiguredPlayer::recoverDec() const 
{
  return ConfiguredPlayer::theRecoverDec;
}

inline double ConfiguredPlayer::recoverMin() const 
{
  return ConfiguredPlayer::theRecoverMin;
}

inline double ConfiguredPlayer::effortDecThr() const 
{
  return ConfiguredPlayer::theEffortDecThr;
}

inline double ConfiguredPlayer::effortDec() const 
{
  return ConfiguredPlayer::theEffortDec;
}

inline double ConfiguredPlayer::effortIncThr() const 
{
  return ConfiguredPlayer::theEffortIncThr;
}

inline double ConfiguredPlayer::effortInc() const 
{
  return ConfiguredPlayer::theEffortInc;
}

inline double ConfiguredPlayer::effortMin() const 
{
  return globalPlayerTypes->effortMin(thePlayerType);
}

inline double ConfiguredPlayer::effortMax() const 
{
  return globalPlayerTypes->effortMax(thePlayerType);
}

inline double ConfiguredPlayer::dashPowerRate() const 
{
  return globalPlayerTypes->dashPowerRate(thePlayerType);
}

inline double ConfiguredPlayer::maxPower() const 
{
  return ConfiguredPlayer::theMaxPower;
}

inline double ConfiguredPlayer::minPower() const 
{
  return ConfiguredPlayer::theMinPower;
}

inline double ConfiguredPlayer::maxMoment() const 
{
  return ConfiguredPlayer::theMaxMoment;
}

inline double ConfiguredPlayer::minMoment() const 
{
  return ConfiguredPlayer::theMinMoment;
}

inline double ConfiguredPlayer::maxNeckMoment() const 
{
  return ConfiguredPlayer::theMaxNeckMoment;
}

inline double ConfiguredPlayer::minNeckMoment() const 
{
  return ConfiguredPlayer::theMinNeckMoment;
}

inline double ConfiguredPlayer::maxNeckAngle() const 
{
  return ConfiguredPlayer::theMaxNeckAngle;
}

inline double ConfiguredPlayer::minNeckAngle() const 
{
  return ConfiguredPlayer::theMinNeckAngle;
}

inline double ConfiguredPlayer::normVisibleAngle() const 
{
  return ConfiguredPlayer::theNormVisibleAngle;
}

inline double ConfiguredPlayer::inertiaMomentOfTurn() const 
{
  return globalPlayerTypes->inertiaMoment(thePlayerType);
}

inline double ConfiguredPlayer::kickableMargin() const 
{
  return globalPlayerTypes->kickableMargin(thePlayerType);
}

inline double ConfiguredPlayer::kickRange() const 
{
  return globalPlayerTypes->kickRange(thePlayerType);
}



inline void ConfiguredPlayer::setStaminaMax(double aStaminaMax) 
{
  ConfiguredPlayer::theStaminaMax=aStaminaMax;
}

inline void ConfiguredPlayer::setRecoverDecThr(double aRecoverDecThr) 
{
  ConfiguredPlayer::theRecoverDecThr=aRecoverDecThr;
}

inline void ConfiguredPlayer::setRecoverDec(double aRecoverDec) 
{
  ConfiguredPlayer::theRecoverDec=aRecoverDec;
}

inline void ConfiguredPlayer::setRecoverMin(double aRecoverMin) 
{
  ConfiguredPlayer::theRecoverMin=aRecoverMin;
}

inline void ConfiguredPlayer::setEffortDecThr(double aEffortDecThr) 
{
  ConfiguredPlayer::theEffortDecThr=aEffortDecThr;
}

inline void ConfiguredPlayer::setEffortDec(double aEffortDec) 
{
  ConfiguredPlayer::theEffortDec=aEffortDec;
}

inline void ConfiguredPlayer::setEffortIncThr(double aEffortIncThr) 
{
  ConfiguredPlayer::theEffortIncThr=aEffortIncThr;
}

inline void ConfiguredPlayer::setEffortInc(double aEffortInc) 
{
  ConfiguredPlayer::theEffortInc=aEffortInc;
}

inline void ConfiguredPlayer::setMaxPower(double aMaxPower) 
{
 ConfiguredPlayer::theMaxPower=aMaxPower;
}

inline void ConfiguredPlayer::setMinPower(double aMinPower) 
{
  ConfiguredPlayer::theMinPower=aMinPower;
}

inline void ConfiguredPlayer::setMaxMoment(double aMaxMoment) 
{
  ConfiguredPlayer::theMaxMoment=aMaxMoment;
}

inline void ConfiguredPlayer::setMinMoment(double aMinMoment) 
{
  ConfiguredPlayer::theMinMoment=aMinMoment;
}

inline void ConfiguredPlayer::setMaxNeckMoment(double aMaxNeckMoment) 
{
  ConfiguredPlayer::theMaxNeckMoment=aMaxNeckMoment;
}

inline void ConfiguredPlayer::setMinNeckMoment(double aMinNeckMoment) 
{
  ConfiguredPlayer::theMinNeckMoment=aMinNeckMoment;
}

inline void ConfiguredPlayer::setMaxNeckAngle(double aMaxNeckAngle) 
{
  ConfiguredPlayer::theMaxNeckAngle=aMaxNeckAngle;
}

inline void ConfiguredPlayer::setMinNeckAngle(double aMinNeckAngle) 
{
  ConfiguredPlayer::theMinNeckAngle=aMinNeckAngle;
}

inline void ConfiguredPlayer::setNormVisibleAngle(double aNormVisibleAngle) 
{
  ConfiguredPlayer::theNormVisibleAngle=aNormVisibleAngle;
}

inline double ConfiguredPlayer::stamina() const 
{
  return theStamina;
}

inline void ConfiguredPlayer::setStamina( double newStamina ) 
{
  theStamina = normalizeStamina( newStamina );
}

inline double ConfiguredPlayer::effort() const 
{
  return theEffort;
}

inline void ConfiguredPlayer::setEffort( double newEffortFactor ) 
{
  theEffort = normalizeEffort( newEffortFactor );
}

inline double ConfiguredPlayer::recovery() const 
{
  return theRecovery;
}

inline void ConfiguredPlayer::setRecovery( double newRecoveryFactor ) 
{
  theRecovery = normalizeRecovery( newRecoveryFactor );
}  

inline double ConfiguredPlayer::maxEffectiveTurnAngle() const  
{ 
  return effectiveTurnAngle( maxMoment() );   
} 

inline void ConfiguredPlayer::setPlayerType(int aPlayerType)
{
  thePlayerType = aPlayerType;
}

    
#endif
