#ifndef _MODELLEDBALL_H_
#define _MODELLEDBALL_H_

#include <stream.h>

#include "ConfiguredBall.h"

#include "../basic/Configuration.h"
#include "../basic/Common.h"
#include "../basic/Vector.h"

/**
  Die interne Repraesentation des Balls im Weltmodell.
  Ein ModelledBall hat zusaetzlich zu den Eigenschaften eines ModelledObject
  das Wissen darueber, von welchem Team er als letzter beruehrt worden ist.
  (Bisher wird diese Eigenschaft aber noch nirgendwo gesetzt und nirgendwo
  abgefragt.)

  {\small {\bf Quelltext}

  [\URL[Header]{src/ModelledBall.h} | [\URL[Implementierung]{src/ModelledBall.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ModelledBall.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/

class ModelledBall : public ConfiguredBall
{
public:


  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Default-Konstruktor und neuer Ball mit Daten.
      @param id Die Id des Balles.
      @param aPos Die Position. Als Default wird unbekannt angenommen.
      @param aSpeed Die Geschwindigkeit. Als Default wird unbekannt angenommen.
      @param aPosConfidence Die Verlaesslichkeit der Position. Als Default wird 
      RANDOM angenommen.
      @param aPosConfidence Die Verlaesslichkeit der Geschwindikeit. Als Default
      wird RANDOM angenommen.
      @param aLastTouchedBy Das Team, von dem der Ball als letztes beruehrt
      worden ist. Default ist UNDEF_TEAM. 
      @param aLastTouchedAtTime Zeit, zu der der Ball zuletzt ber\"uhrt wurde.
      Default ist -1. */
    ModelledBall( int id = UNDEF_ID,
                  const Vector& aPos = Vector(), 
                  const Vector& aSpeed = Vector(),
                  int aLastTouchedBy = UNDEF_TEAM_ID,
                  int aLastTouchedAtTime = -1,
		  const Confidence& aPosConfidence = Confidence(),
		  const Confidence& aSpeedConfidence = Confidence() );

    /** Modellierung eines wahrgenommenen Balls.
      Dabei werden die Positions- und Geschwindigkeitsvektoren NICHT in das
      absolute Koordinatensystem \"ubertragen, sondern einfach \"ubernommen.
      Die Transformation ist Aufgabe des Weltmodells.
      @param aMovingObject Der gesehene Ball.
      @param aPosConfidence Die Verlaesslichkeit der Position. Als Default
      wird RANDOM angenommen.
      @param aPosConfidence Die Verlaesslichkeit der Geschwindikeit. Als
      Default wird RANDOM angenommen.
      @param aLastTouchedBy Das Team, von dem der Ball als letztes beruehrt
      worden ist. Default ist UNDEF_TEAM_ID.
      @param aLastTouchedAtTime Zeit, zu der der Ball zuletzt ber\"uhrt wurde.
      Default ist -1. */
    ModelledBall( const MovingObject& aMovingObject,
		  int aLastTouchedBy = UNDEF_TEAM_ID,
		  int aLastTouchedAtTime = -1,
                  const Confidence& aPosConfidence = Confidence(),
	          const Confidence& aSpeedConfidence = Confidence() );


  //@}
    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
    StableObject* copy() const;

  /** @name Werte abfragen */
  //@{
    /** Liefert Team, das zuletzt den Ball beruehrt hat. */
    int lastTouchedBy() const;

    /** Liefert die Zeit, zu der der Ball zuletzt ber\"uhrt wurde. */
    int lastTouchedAtTime() const;
 //@}

    /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal.
      Es werden ID, Position, Geschwindigkeit, Zuverlaessigkeitswerte,
      LastTouchedBy und LastTouchedAtTime ausgegeben. */
    Ath98String toString() const;

  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const ModelledBall& );

  /** kompakte Ausgabe auf einen Stream fr constant Log */
    void toStampStream( ostream& os) const;

  /** @name Werte setzen */
  //@{         
    /** Setzt Team, das zuletzt den Ball beruehrt hat, neu. */
    void setLastTouchedBy( int  newLastTouchedBy );

    /** Setzt die Zeit, zu der der Ball zuletzt ber\"uhrt wurde. */
    void setLastTouchedAtTime( int  newLastTouchedAtTime );
  //@}
    
  
  /** @name Simulation */
  //@{
    /** Fuehrt Simulationsschritte aus. 
      Dabei wird die Geschwindigkeit auf die Position addiert und die
      Geschwindigkeit mit dem Verlangsamungsfaktor multipliziert. Die
      Verlaesslichkeit wird herabgesetzt. */
    void simulateSteps( unsigned int simSteps = 1 );

    /** Beschleunigt. 
      Wie stark sich die Kraft auf die Beschleunigung auswirkt, haengt vom
      Kraftfaktor (powerRate), von einem Objektbeschleunigungsfaktor
      (maxPowerRate) und von der Position des kickenden Spielers (byPlayer) ab.
      
      Durch Aufruf von accelerate werden nur Werte gesetzt. Um die Aktion
      tats\"achlich zu simulieren, mu\3 anschlie\3end \Ref{simulateSteps}
      aufgerufen werden.
      @param power Die Kraft, mit der beschleunigt wird. 
      @param angle Die Richtung, relativ zur Blickrichtung des Spielers
      byPlayer, in die beschleunigt werden soll.
      @param byPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @exception ModelledBallException bei undefinierter Ballgeschwindigkeit */
    void accelerate( double power, double angle,
		    const ModelledPlayer& byPlayer );
  //@}

  /** @name Hilfsfunktionen zur Kick-Berechnung */
  //@{
    /** Ueberprueft, ob der Ball im Kick-Bereich ist.
      @return YES, wenn ja, sonst nein */
      bool isInKickRangeOfPlayer( const ModelledPlayer& aPlayer ) const;
      
    /** Es wird die tatsaechlich verfuegbare Kraft
      (effectiveKickPower) ausgerechnet.
      @param power Die Kraft, mit der man den Ball kicken will.
      @param kickingPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @return effectiveKickPower Die Kraft, mit der der Ball tatsaechlich
      beschleunigt wird.
      @exception KickImpossibleException falls Ball nicht im Kick-Bereich (zu 
      weit weg) */

  /** Id des Teams, das den Ball als letztes beruehrt hat. */
  int theLastTouchedBy;  

  /** Zeit zu der der Ball zuletzt ber\"uhrt wurde. */
  int theLastTouchedAtTime;
};

inline int ModelledBall::lastTouchedBy() const
{
    return theLastTouchedBy;
}


inline int ModelledBall::lastTouchedAtTime() const
{
    return theLastTouchedAtTime;
}


inline ostream& operator<< (ostream& os, const ModelledBall& ball )
{
  os << 
    "( ID: " << ball.id()  <<
    ", Pos: " << ball.pos() << 
    ", PosConfidence"  << ball.posConfidence() <<
    ", Speed: " << ball.speed() << 
    ", SpeedConfidence" << ball.speedConfidence() <<
    ", LastTouchedBy: " << ball.theLastTouchedBy <<
    ", LastTouchedAtTime: " << ball.theLastTouchedAtTime <<
    " )" ;
  return os;
}

inline void ModelledBall::toStampStream( ostream& os) const {
  os << 
    " " << pos().x() << 
    " " << pos().y() <<
    " " << speed().x() <<
    " " << speed().y();
}

#endif
