#include <math.h>

#include "ModelledPlayer.h"

#include "Field.h"
#include "Situation.h"
#include "WorldModelInterface.h"

#include "../basic/Configuration.h"
#include "../basic/Log.h"
#include "../basic/Action.h"


/** 
 * Aktueller Stand 
 * 
 *   $RCSfile: ModelledPlayer.cc,v $ 
 *   $Revision: 1.2 $ 
 *   $Date: 2001/08/02 08:09:42 $ 
 *   $Author: rami $  
 * 
 */ 

const int ModelledPlayer::MAX_STEPS_TO_BALL = 100;  
 
ModelledPlayer::ModelledPlayer(int aId,
			                         const Vector& aPos,  
			                         const Vector& aSpeed,
			                         double aFaceDir,
			                         double aBodyDir, 
                               int aViewWidth,
                               int aViewQuality,
                               double stamina,
                               double effort, 
                               double recovery,
                               bool isGoalie,
                               const Confidence& aPosConfidence,
	                             const Confidence& aSpeedConfidence,
 	                             const Confidence& aFaceDirConfidence,
 	                             const Confidence& aBodyDirConfidence,
			                         const Confidence& aStaminaConfidence,
			                         const Confidence& aTeamIdConfidence,
			                         const Confidence& aPlayerIdConfidence
			                          )
  : ConfiguredPlayer( aId, aPos, aSpeed, stamina, effort, recovery,
                      aPosConfidence, aSpeedConfidence ),
    theFaceDir( aFaceDir ), 
    theBodyDir( aBodyDir ),
    theFaceDirConfidence( aFaceDirConfidence ), 
    theBodyDirConfidence( aBodyDirConfidence ), 
    theStaminaConfidence( aStaminaConfidence ), 
    theTeamIdConfidence( aTeamIdConfidence ), 
    thePlayerIdConfidence( aPlayerIdConfidence ), 
    isTheGoalie( isGoalie )
{
  setViewWidth( aViewWidth );
  setViewQuality( aViewQuality );
  if ( ModelledPlayer::isSetOnce == NO ) {
    setClassVars();  
  } 
}

ModelledPlayer::ModelledPlayer(const MovingPlayer& aMovingPlayer,
			                         int aViewWidth,
			                         int aViewQuality,
			                         double stamina, 
			                         double effort,
			                         double recovery,
			                         const Confidence& aPosConfidence,
                  			       const Confidence& aSpeedConfidence,
			                         const Confidence& aFaceDirConfidence,
			                         const Confidence& aBodyDirConfidence,
                               const Confidence& aStaminaConfidence,
			                         const Confidence& aTeamIdConfidence,
			                         const Confidence& aPlayerIdConfidence )
  : ConfiguredPlayer( aMovingPlayer.id(),
		                  aMovingPlayer.pos(),
		                  aMovingPlayer.speed(),
		                  stamina, effort, recovery,
		                  aPosConfidence, aSpeedConfidence ),
    theFaceDir( aMovingPlayer.faceDir() ),
    theBodyDir( aMovingPlayer.bodyDir() ),
    theFaceDirConfidence( aFaceDirConfidence ),
    theBodyDirConfidence( aBodyDirConfidence ),
    theStaminaConfidence( aStaminaConfidence ),
    theTeamIdConfidence( aTeamIdConfidence ), 
    thePlayerIdConfidence( aPlayerIdConfidence ), 
    isTheGoalie( aMovingPlayer.isGoalie() )
{
  setViewWidth( aViewWidth );
  setViewQuality( aViewQuality );
  if ( isSetOnce == NO ) {
    setClassVars();
  }
}


StableObject* ModelledPlayer::copy() const
{
  return new ModelledPlayer( *this );
}

double ModelledPlayer::maxEffectiveDashPower() const
{
  INFO( Log::MODELLED_OBJECTS  ,  4  ,  idString() , globalLog->infoStream
    << " ENTER: ConfiguredPlayer::maxEffectiveDashPower\n"  ) ;

  // wie stark kann ich berhaupt dashen (wegen Stamina)
  double power = effectiveDashPower( maxPower() );

//cerr << "effectiveDashPower( maxPower() ) : " << power << endl;

  // Beschleunigungsvektor (dashPowerRate() aus der server.conf (2001) sollte 0.006 sein)
  Vector acceleration( dashPowerRate() * power, bodyDir(), POLAR, RADIAN );

//cerr << "acceleration = " << acceleration.length() << endl;
//cerr << "speed = " << speed().length() << endl;

  if( (speed() + acceleration).length() > maxSpeed() ) {
    // bei momentaner Beschleunigung wrde unser speed > maxSpeed werden
    // --> Vergeudung von Stamina, da der speed auf maxSpeed zurckgesetzt wird

    // bei der power mte speed und acceleration = maxSpeed sein
    power = signum( power ) * (maxSpeed() - speed().length()) / dashPowerRate();
  }

  INFO( Log::MODELLED_OBJECTS  ,  4  ,  idString() , globalLog->infoStream
    << " LEAVE: ConfiguredPlayer::maxEffectiveDashPower\n"  ) ;
//cerr << "resultPower = " << power << endl;
  return power;
}

void ModelledPlayer::setVisibleAngle( int viewWidth )
{
  switch( viewWidth ) {
  case VIEW_WIDTH_NARROW:
    theVisibleAngle = normVisibleAngle() / 2;
    break;
  case VIEW_WIDTH_NORMAL:
    theVisibleAngle = normVisibleAngle();
    break;
  case VIEW_WIDTH_WIDE:
    theVisibleAngle = normVisibleAngle() * 2;
    break;
  default:
    throwException( ModelledPlayerException( "viewWidth undef" ) );
  }
}

void ModelledPlayer::simulateSteps( unsigned int simSteps )
{ 
	simSteps = max((int) simSteps, 0);
	simSteps = min((int) simSteps, 100);
	
  while( simSteps > 0 ) {   
    setPos( pos() + speed() );  
    setSpeed( decay() * speed() );  
    thePosConfidence.decrease();
    theSpeedConfidence.decrease();
    theFaceDirConfidence.decrease();
    theBodyDirConfidence.decrease();
    recover( 1 );
    --simSteps;
  } 
}

 
void ModelledPlayer::accelerate(double power)  
{ 
  power = normalizeDashPower( power );

  if( speedConfidence().reliability() > 0.0 + DEFAULT_COMPARISON_ACCURACY ) {
    //cerr << "ModelledPlayer::speedConfidence = 0\n";
  }
  double effPower = effectiveDashPower( power );
  Vector myAcceleration = Vector( dashPowerRate() * effPower, bodyDir(), POLAR, RADIAN );
  setSpeed( speed() + myAcceleration );
  if( speed().length() > maxSpeed() ) { 
      setSpeed( Vector( maxSpeed(), speed().angle(), POLAR, RADIAN ) );
  }

  theStamina -= (power > 0) ? (power) : (-2.0 * power);
  theStamina = max( 0.0, theStamina );
} 

 
void ModelledPlayer::turn( double angle )  
{ 
  INFO( Log::MODELLED_OBJECTS  ,  2  ,  idString() , globalLog->infoStream 
    << " ENTER ModelledPlayer::turn \n"  ) ; 

  angle = effectiveTurnAngle(angle);
  double relAngle = ( angle > Ath98_PI ) ? (2.0 * Ath98_PI) - angle : (-1.0 * angle);
  theFaceDir = normalizeAngle( theFaceDir - relAngle );  
  theBodyDir = normalizeAngle( theBodyDir - relAngle );  

  INFO( Log::MODELLED_OBJECTS  ,  2  ,  idString() , globalLog->infoStream 
    << " LEAVE ModelledPlayer::turn \n" ) ; 
} 


double ModelledPlayer::angleOfFaceTo( const Vector& point ) const  
{ 
  Vector relativePos = point - pos();  
  if( faceDir() != UNDEF_DOUBLE ) {
    return normalizeAngle( relativePos.angle() - faceDir() );  
  } 
  else{ 
//    throwException( ModelledPlayerException( "ModelledPlayer has undefined faceDir" ) );
    cerr << "ModelledPlayer::angleOfFaceTo    ModelledPlayer has undefined faceDir" << endl;
    return 0.0;
  } 
} 


double ModelledPlayer::angleOfBodyTo( const Vector& point ) const  
{
  Vector relativePos = point - pos();
  if( bodyDir() != UNDEF_DOUBLE ) { 
    return normalizeAngle( relativePos.angle() - bodyDir() );
  } 
  else{ 
//    throwException( ModelledPlayerException( "ModelledPlayer has undefined bodyDir" ) );
    cerr << "ModelledPlayer::angleOfBodyTo    ModelledPlayer has undefined bodyDir" << endl;
    return 0.0;

  } 
} 


bool ModelledPlayer::isInFieldOfView( const StableObject& object ) const  
{ 
  double angle = angleOfFaceTo( object.pos() );  
  // relAngle braucht nicht negative zu sein fr den Vergleich --> Operation gespart
  double relAngle = ( angle > Ath98_PI ) ? (2.0 * Ath98_PI) - angle : angle;

  if( relAngle < visibleAngle() / 2.0 ) { 
    return YES;  
  } 
  else { 
    return NO;  
  } 
} 


bool ModelledPlayer::isInFieldOfFeel( const StableObject& object ) const 
{ 
  if( (object.pos() - pos()).length() <= globalConf->visibleDistance ) { 
    return YES;  
  } 
  else { 
    return NO;  
  } 
} 


bool ModelledPlayer::isInFieldOfSense( const StableObject& object ) const 
{ 
  return ( isInFieldOfView( object ) || isInFieldOfFeel( object ) );  
} 


 
int ModelledPlayer::team() const 
{ 
  if ( id() >= MIN_TEAMMATE_ID && id() <= MAX_TEAMMATE_ID ) { 
    return OUR_TEAM_ID;  
  } 
  else if( id() >= MIN_OPPONENT_ID && id() <= MAX_OPPONENT_ID ) { 
    return THEIR_TEAM_ID;  
  } 
  else { 
    return UNDEF_TEAM_ID;  
  } 
} 

   
bool ModelledPlayer::isPlayerNumberKnown() const 
{ 
  if ( (id() == UNDEF_OPPONENT_ID) ||
       (id() == UNDEF_TEAMMATE_ID) ||
       (id() == UNDEF_PLAYER_ID) ) {
    return NO;
  }
  else {
    return YES;
  }
}

 
Vector ModelledPlayer::newCoordinates(const Vector &ballPos, const Vector &playerPos, const Vector &ballSpeed) const 
{ 
  // Formel, um die neuen Koordinaten zu errechnen (lt. Stoeker 1993, S.307f.) 
  // 1. Schritt: Paralellverschiebung des Spielerpunktes in Richtung des neuen Ursprungs, 
  //             was der Ballposition entspricht 
  // 2. Schritt: Passive Drehung des Koordinatensystems um den Winkel, den der 
  //             Ballgeschwindigkeitsvektor zur x-Achse hat. 
 
  // 1. Schritt 
  Vector helpVec = playerPos - ballPos;  
  // Zwischenwerte 
  double angle = ballSpeed.angle();  
  double cosFromAngle = cos( angle );  
  double sinFromAngle = sin( angle );  
  // 2. Schritt 
  Vector newPos( helpVec.x() * cosFromAngle + helpVec.y() * sinFromAngle, 
		 helpVec.y() * cosFromAngle + helpVec.x() * sinFromAngle);  
   
//   double radius = ( playerPos - ballPos ).length();  
//   double speedAngle = ballSpeed.angle();  
//   double playerBallSpeedAngle = ( ballSpeed - playerPos ).angle();  
   
//   double playerAngle = playerPos.angle();  
 
//   double angle;  
 
//   cout << "Winkel:" << endl;  
//   cout << "radius: " << radius << endl;  
//   cout << "speed : " << speedAngle << endl;  
//   cout << "player: " << playerAngle << endl;  
//   cout << "pl-spedd: " << playerBallSpeedAngle << endl;  
   
//   angle =  fabs( speedAngle - playerBallSpeedAngle );  
//   if ( angle > 2*Ath98_PI ){ 
//     angle -= 2*Ath98_PI;  
//   } 
//   cout << "angle : " << angle << endl;  
   
//   // Gleich den Vektor in Polarkoordination anlegen 
//   Vector newPos( radius, angle, POLAR );  
//   cout << "newPos: " << newPos.toString() << endl;  
//   newPos.setCartesian( newPos.x(),fabs(newPos.y()) );  
  //cout << "newPos0:" << newPos.toString() << endl;  
  if( newPos.y() < 0 ){ 
    newPos.setCartesian( newPos.x(), fabs( newPos.y() ) );  
  } 
  //cout << "newPos0:" << newPos.toString() << endl;  
  return newPos;  
} 

 
 Ath98String ModelledPlayer::toString() const
{
  Ath98String output;

  output = Ath98String("( ID: ") + Ath98String( id() ) +
    Ath98String(", Id") + playerIdConfidence().toString() +
    Ath98String(", Team") + teamIdConfidence().toString() +
    Ath98String(", IsGoalie: ") + Ath98String( isGoalie() ) +
    Ath98String(", Pos:") + pos().toString() +
    Ath98String(", Pos") + posConfidence().toString() +
    Ath98String(", Speed:") + speed().toString() +
    Ath98String(", Speed") + speedConfidence().toString() +
    Ath98String(", FaceDir: ") + Ath98String(degFromRadiant(faceDir())) +
    Ath98String(", FaceDir") + faceDirConfidence().toString() +
    Ath98String(", BodyDir: ") + Ath98String(degFromRadiant(bodyDir())) +
    Ath98String(", BodyDir") + bodyDirConfidence().toString() +
    Ath98String(", VisibleAngle: ") + Ath98String(degFromRadiant(visibleAngle())) +
    Ath98String(", Stamina: ") + Ath98String(stamina()) +
    Ath98String(", Recovery: ") + Ath98String(recovery()) +
    Ath98String(", Effort: ") + Ath98String(effort()) +
    Ath98String(", StaminaValues") + staminaConfidence().toString() +
    Ath98String(" )");
 
  return output;
}                                                                                                                      


