#ifndef _STABLEOBJECT_H_
#define _STABLEOBJECT_H_
 
#include "../basic/Vector.h"
#include <stream.h>




/**
  Die Objekt-Repraesentation eines in der Aussenwelt vorkommenden Objektes.
  Solche Objekte sind Flaggen, Linien, Tore, Spieler und der Ball.

  Ein StableObject hat eine Position (ein 2D-Vektor) und eine Id, in der
  kodiert ist, um welchen Typ von Objekt es sich handelt, und welches der
  Objekte dieses Typs das StableObject repraesentiert.

  StableObject ist gleichzeitig die Basisklasse fuer weitere spezialisierte
  Repraesentationsklassen wie MovingObject oder MovingPlayer.

  StableObjects kommen vor allem in der Wahrnehmung vor. Sensors erzeugt sie.
  Das Weltmodell erstellt aus ihnen eine neue Situation.

  Die in einem StableObject aufgenommenen Daten sind immer im mathematischen
  Koordinatensystem. Das Weltmodell muss sie also bei der Berechnung der
  absoluten Koordinaten nicht mehr umwandeln. 

  Ob die Daten als relativ zum Agenten und seiner Blickrichtung oder als
  absolut (d.h. relativ zur Spielfeldmitte bei zum Gegner zeigender positiver
  x-Achse) interpretiert werden, ist nicht festgelegt. \Ref{Sensors} setzt die
  Daten relativ zum Agenten, \Ref{WorldModel} setzt die Daten in Subklassen
  (\Ref{ModelledObject} und davon abgeleitet \Ref{ModelledPlayer} und
  \Ref{ModelledBall} von StableObject absolut.

  {\small {\bf Quelltext}

  [\URL[Header]{src/StableObject.h} | [\URL[Implementierung]{src/StableObject.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: StableObject.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $
  *
  **************************************************
  \end{verbatim} }*/
class StableObject{
public:

  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Default-Konstruktor.
      Id ist UNDEF_ID und die Position ist der undefinierte Vektor.*/
    StableObject();

    /**  Initialisierung mit einem Vektor.
      @param anId die Id des Objektes. Moegliche Werte sind in \Ref{Objekt-Ids}
      aufgelistet.
      @param aVector die Position des Objektes. Das Koordinatensystem h\"angt
      vom Verwendungszweck ab.*/
    StableObject( int anId, const Vector& aPoint );


    /*@param Ein Zeiger vom Typ (StableObject *) auf den neu allozierten
      Speicher. */
    StableObject* copy() const;

    /** Virtueller Destruktor. */
    ~StableObject();
  //@}

  /** @name Werte setzen */
  //@{
    /** Setzt die Position. */
    void setPos(const Vector& newPos);

    /** Setzt die ObjectId. */
    void setId(int);
  //@}

  /** @name Werte abfragen */
  //@{
    /** Liefert die Position des Objekts.
      Das Koordinatensystem haengt von der Verwendung ab. */
    const Vector& pos() const;

    /** Liefert die Id des Objekts. */
    int id() const;

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
    // leda_string idString() const;

    /** Test auf Gueltigkeit der Id. */
    bool isValid() const;

    /** Prueft, ob das Objekt komplett definiert ist..
      @return YES falls die Id verschieden von UNDEF_ID ist und die Position
        complete ist, sonst NO */
    bool isComplete() const;

  //@}
    /** Ausgabe der Beschreibung auf einen Stream.
      Standardm\"a\3ig wird auf stdout ausgegeben. Der auszugebende
      Ath98String wird von der Methode \Ref{toString}() erzeugt.
      Am einfachsten kann man das Objekt durch Umleitung nach cout ausgeben. */
    void printOn (ostream& = cout) const;

  /** Liefert einen String mit Klassenbezeichnung zurueck. */
  string idString() const;


  /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const StableObject& );

   /** Beschreibung als Ath98String fuer die Ausgabe auf einem Terminal.
      Der Ath98String sollte kein abschlie\3sendes newline enthalten. Die Ausgabe
      sollte in () eingschlossen sein und als erstes den Klassennamen
      enthalten.
      Die Methode mu\3 in der abgeleiteten Klasse \"uberschrieben werden.*/
    Ath98String toString() const;

    /** Ausgabe auf den InfoStream ohne Benutzung von Ath98String
     wird vom Operator << benutzt und sollte in jeder abgeleiteten Klasse
    redefiniert werden. */
    void toInfoStream(ostream& os) const;

 protected :
   /** String der die Kurzkennung der Klasse enthlt */
  string theIdString;

 private:
  /** Die Id des Objekts. Das ist KEINE Id im Sinne objektorientierter
    Programmierung. Hier stehen vielmehr Zahlen, die fuer bestimmte Objekte in
    der Aussenwelt stehen. */
  int theId;


  /** Die Position. Hier koennen nur relative Koordinaten enthalten
    sein. Winkel sind immer im Bogenmass, das Koordinatensystem ist das
    mathematische. */
  Vector thePos;
};

inline const Vector& StableObject::pos() const
{
  return thePos;
}

inline int StableObject::id() const
{
  return theId;
}

inline void StableObject::setPos(const Vector& newPos)
{
  thePos = newPos;
}

inline void StableObject::setId(int newId)
{
  theId = newId;
}

inline bool StableObject::isValid() const
{
  return (theId != UNDEF_ID);
}

inline bool StableObject::isComplete() const
{
  return (theId != UNDEF_ID && thePos.isValid() );
}

inline string StableObject::idString() const
 {
   return theIdString;
 }

inline ostream& operator<< (ostream& os, const StableObject&  StableObject )
{
  os <<
    " id: "  << StableObject.id() <<
    ", position: " << StableObject.pos() << " )" ;
  return os;
}

#endif
